# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

def filtered_field(field, filter):
    for f in filter:
        yield field[f]


class FilteredField:
    def __init__(self, field, filter):
        self.field = field
        self.filter = filter
        self.dtype = self.field.dtype if isinstance(self.field, np.ndarray) else None

    def __getitem__(self, item):
        return self.field[self.filter[item]]

    def __len__(self):
        return len(self.filter)





from . import age_from_year_of_birth, analytics, assessment_merge, convert_old_assessments, \
    covid_test, inconsistent_symptoms, nat_medicine_model, numpy_buffer, temperature, \
    weight_height_bmi



#!/usr/bin/env python

from collections import defaultdict
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta
import time

import numpy as np
import h5py
import pandas as pd

from hystore.core import exporter, persistence, utils
from hystore.core.persistence import DataStore
from hystore.processing.nat_medicine_model import nature_medicine_model_1
from hystore.processing.effective_test_date import effective_test_date


"""
avenues to explore:
 * remove other health fieds than health_worker_combined
 * geocode stuff gone
 * imd stuff in
 * healthcare stuff in
 * assessment_frequency in
 * imd for uk only
"""



# patient fields to export
patient_fields_to_export =\
    ['id', 'created_at', 'created_at_day', 'updated_at', 'updated_at_day', 'country_code',
     'version', 'a1c_measurement_mmol', 'a1c_measurement_mmol_valid', 'a1c_measurement_percent',
     'a1c_measurement_percent_valid', 'activity_change', 'age', 'alcohol_change',
     'already_had_covid', 'always_used_shortage', 'assessment_count', 'blood_group', 'bmi',
     'bmi_clean', 'cancer_type', 'contact_health_worker', 'daily_assessment_count',
     # 'diabetes_diagnosis_year', 'diabetes_diagnosis_year_valid', 'diabetes_oral_biguanide',
     # 'diabetes_oral_dpp4', 'diabetes_oral_meglitinides', 'diabetes_oral_other_medication',
     # 'diabetes_oral_sglt2', 'diabetes_oral_sulfonylurea', 'diabetes_oral_thiazolidinediones',
     # 'diabetes_treatment_basal_insulin', 'diabetes_treatment_insulin_pump',
     # 'diabetes_treatment_lifestyle', 'diabetes_treatment_none',
     # 'diabetes_treatment_other_injection', 'diabetes_treatment_other_oral',
     # 'diabetes_treatment_pfnts', 'diabetes_treatment_rapid_insulin',
     'diabetes_type',
     'diabetes_type_other',
     # 'diabetes_uses_cgm',
     'diet_change', 'does_chemotherapy', 'ethnicity',
     'ever_had_covid_test', 'first_assessment_day', 'gender', 'has_asthma', 'has_cancer',
     'has_diabetes', 'has_eczema', 'has_hayfever', 'has_heart_disease', 'has_kidney_disease',
     'has_lung_disease', 'has_lung_disease_only', 'have_used_PPE',
     'have_worked_in_hospital_care_facility', 'have_worked_in_hospital_clinic',
     'have_worked_in_hospital_home_health', 'have_worked_in_hospital_inpatient',
     'have_worked_in_hospital_other', 'have_worked_in_hospital_outpatient',
     'have_worked_in_hospital_school_clinic', 'healthcare_professional', 'height_cm',
     'height_cm_clean', 'height_cm_valid', 'help_available', 'housebound_problems',
     'ht_combined_oral_contraceptive_pill', 'ht_depot_injection_or_implant',
     'ht_hormone_treatment_therapy', 'ht_mirena_or_other_coil', 'ht_none',
     'ht_oestrogen_hormone_therapy', 'ht_pfnts', 'ht_progestone_only_pill',
     'health_worker_with_contact', 'ht_testosterone_hormone_therapy', 'imd_decile', 'imd_rank',
     'interacted_patients_with_covid', 'interacted_with_covid', 'is_carer_for_community',
     'is_pregnant', 'is_smoker', 'last_asked_level_of_isolation',
     'last_asked_level_of_isolation_day', 'last_assessment_day', 'lifestyle_version',
     'limited_activity', 'mobility_aid', 'need_inside_help', 'need_outside_help', 'needs_help',
     'never_used_shortage',
     # 'past_symptom_abdominal_pain', 'past_symptom_anosmia',
     # 'past_symptom_chest_pain', 'past_symptom_delirium', 'past_symptom_diarrhoea',
     # 'past_symptom_fatigue', 'past_symptom_fever', 'past_symptom_hoarse_voice',
     # 'past_symptom_persistent_cough', 'past_symptom_shortness_of_breath',
     # 'past_symptom_skipped_meals', 'past_symptoms_changed', 'past_symptoms_days_ago',
     # 'past_symptoms_days_ago_valid',
     'period_frequency', 'period_status', 'period_stopped_age',
     'period_stopped_age_valid', 'pregnant_weeks', 'pregnant_weeks_valid', 'race_is_other',
     'race_is_prefer_not_to_say', 'race_is_uk_asian', 'race_is_uk_black', 'race_is_uk_chinese',
     'race_is_uk_middle_eastern', 'race_is_uk_mixed_other', 'race_is_uk_mixed_white_black',
     'race_is_uk_white', 'race_is_us_asian', 'race_is_us_black', 'race_is_us_hawaiian_pacific',
     'race_is_us_indian_native', 'race_is_us_white', 'race_other', 'reported_by_another',
     'same_household_as_reporter', 'se_postcode', 'smoked_years_ago', 'smoked_years_ago_valid',
     'smoker_status', 'snacking_change', 'sometimes_used_shortage', 'still_have_past_symptoms',
     'takes_any_blood_pressure_medications', 'takes_aspirin',
     'takes_blood_pressure_medications_pril', 'takes_blood_pressure_medications_sartan',
     'takes_corticosteroids', 'takes_immunosuppressants', 'test_count', 'unwell_month_before',
     # 'vs_asked_at', 'vs_asked_at_day', 'vs_garlic', 'vs_multivitamins', 'vs_none', 'vs_omega_3',
     # 'vs_pftns', 'vs_probiotics', 'vs_vitamin_c', 'vs_vitamin_d', 'vs_zinc',
     'weight_change',
     'weight_change_kg', 'weight_change_kg_valid', 'weight_change_pounds',
     'weight_change_pounds_valid', 'weight_kg', 'weight_kg_clean', 'weight_kg_valid', 'zipcode']


def supplement_paper_processing(ds, src_data, dest_data):
    src_ptnts = src_data['patients']
    src_asmts = src_data['assessments']
    src_tests = src_data['tests']

    a_cats_ = ds.get_reader(src_asmts['created_at'])[:]
    start_timestamp = a_cats_.min()
    end_timestamp = a_cats_.max()

    supplement_fields = ('vs_garlic', 'vs_multivitamins', 'vs_none', 'vs_omega_3',
                         'vs_pftns', 'vs_probiotics', 'vs_vitamin_c', 'vs_vitamin_d', 'vs_zinc')

    verbose = True
    def filter_status(verbose, name, field, indent=0):
        if verbose:
            preamble = ''.join([" "] * indent)
            print('{}filter after {}: {}/{}'.format(
                preamble, name, np.count_nonzero(field), np.count_nonzero(field == False)))

    # initial patient filtering / flattening
    # --------------------------------------
    if 'filtered_patients' not in dest_data:
        with utils.Timer("initial patient filtering / flattening", new_line=True):
            flt_ptnts = dest_data.create_group('filtered_patients')

            patient_filter = np.ones(len(ds.get_reader(src_ptnts['id'])), dtype=np.bool)
            print("  Initial patient count; {} patients included".format(len(patient_filter)))

            patient_filter = np.ones(len(ds.get_reader(src_ptnts['id'])), dtype=np.bool)
            for k in ('16_to_90_years', '40_to_200_kg', '110_to_220_cm', '15_to_55_bmi'):
                patient_filter = patient_filter & ds.get_reader(src_ptnts[k])[:]
                filter_status(verbose, "{}".format(k), patient_filter, 2)


            # get gender filter
            ptnt_gender = ds.get_reader(src_ptnts['gender'])[:]
            gender_filter = (ptnt_gender == 1) | (ptnt_gender == 2)
            patient_filter = patient_filter & gender_filter
            del gender_filter
            filter_status(verbose, 'gender', patient_filter, 2)

            # get filter for people who were asked questions
            vs_asked_filter = ds.get_reader(src_ptnts['vs_asked_at'])[:] != 0
            patient_filter = patient_filter & vs_asked_filter
            del vs_asked_filter
            filter_status(verbose, 'vs_asked_at', patient_filter, 2)


            # get filter for people with assessments
            has_assessments_filter = ds.get_reader(src_ptnts['daily_assessment_count'])[:] > 0
            patient_filter = patient_filter & has_assessments_filter
            del has_assessments_filter
            filter_status(verbose, 'daily_assessment_count', patient_filter, 2)

            # filter and flatten patient fields
            # ---------------------------------

            for k in patient_fields_to_export:
                with utils.Timer("  filtering {}".format(k)):
                    if k == 'gender':
                        r = ds.get_reader(src_ptnts[k])
                        ds.apply_filter(patient_filter, r[:] - 1,
                                        r.get_writer(flt_ptnts, k))
                    else:
                        r = ds.get_reader(src_ptnts[k])
                        ds.apply_filter(patient_filter, r, r.get_writer(flt_ptnts, k))

            # filter and flatten supplement fields
            for s in supplement_fields:
                with utils.Timer("  filtering {}".format(s)):
                    reader = ds.get_reader(src_ptnts[s])
                    writer = ds.get_numeric_writer(flt_ptnts, s, 'int8')
                    filtered = ds.apply_filter(patient_filter, reader)
                    filtered = filtered.astype('int8')
                    filtered = np.where((filtered - 1) < 0, 0, filtered - 1)
                    writer.write(filtered)

            for s in supplement_fields:
                print("  {}".format(s), utils.build_histogram(ds.get_reader(flt_ptnts[s])[:]))

            print("applying asked about supplements filter; {} patients included".format(
                np.count_nonzero(patient_filter == True)
            ))
            patient_indices = np.arange(len(patient_filter))[patient_filter != 0]
            ds.get_numeric_writer(flt_ptnts, 'patient_indices', 'int64').write(patient_indices)

    else:
        print('  filtered patients already calculated')
        flt_ptnts = dest_data['filtered_patients']

    # initial assessment filtering / flattening
    # -----------------------------------------
    if 'filtered_assessments' not in dest_data.keys():
        with utils.Timer("filtering / flattening assessments", new_line=True):
            flt_asmts = dest_data.create_group('filtered_assessments')

            with utils.Timer(f"filtering out orphaned assessments"):
                asmt_not_orphaned_filter =\
                    persistence.foreign_key_is_in_primary_key(
                        ds.get_reader(flt_ptnts['id']), ds.get_reader(src_asmts['patient_id']))

            asmt_filter = asmt_not_orphaned_filter

            symptom_operators = ('persistent_cough', 'skipped_meals', 'loss_of_smell',
                                 'fatigue')
            symptom_thresholds = {s: 2 for s in symptom_operators}
            symptom_thresholds['fatigue'] = 3
            for s in symptom_operators:
                with utils.Timer(f"flattening and filtering {s}"):
                    reader = ds.get_reader(src_asmts[s])
                    writer = ds.get_numeric_writer(flt_asmts, s, 'uint8')
                    writer.write(
                        ds.apply_filter(asmt_filter, reader[:] >= symptom_thresholds[s]))

            # filter assessment_ids
            asmt_core_fields = ('id', 'patient_id', 'created_at', 'updated_at',
                                'tested_covid_positive')

            for k in asmt_core_fields:
                with utils.Timer(f"filtering '{k}'"):
                    reader = ds.get_reader(src_asmts[k])
                    ds.apply_filter(asmt_filter, reader,
                                    reader.get_writer(flt_asmts, k))

            with utils.Timer("converting and filtering 'location' -> 'hospitalization' field"):
                asmt_locations = ds.get_reader(src_asmts['location'])
                ds.apply_filter(asmt_filter, asmt_locations[:] > 1,
                                ds.get_numeric_writer(flt_asmts, 'hospitalization', 'uint8'))

            with utils.Timer('flattening treatment field'):
                asmt_treatment = ds.get_reader(src_asmts['treatment'])
                ds.apply_filter(asmt_filter, asmt_treatment[:] > 1,
                                ds.get_numeric_writer(flt_asmts, 'treatment', 'uint8'))

            with utils.Timer('flattening health_status field'):
                ds.apply_filter(asmt_filter, ds.get_reader(src_asmts['health_status'])[:] < 2,
                                ds.get_numeric_writer(flt_asmts, 'healthy', 'bool'))
    else:
        print('filtered assessments already calculated')
        flt_asmts = dest_data['filtered_assessments']

    # calculate initially healthy flag and join to patients
    # -----------------------------------------------------
    if 'initially_healthy' not in flt_ptnts:
        with utils.Timer('determining initially unhealthy patients', new_line=True):
            a_pids_ = ds.get_reader(flt_asmts['patient_id'])[:]
            a_hlts_ = ds.get_reader(flt_asmts['healthy'])[:]
            print("  ", sorted(utils.build_histogram(a_hlts_)))
            asmt_spans = ds.get_spans(a_pids_)
            first_asmt_indices = ds.apply_spans_index_of_first(asmt_spans)
            f_a_pids_ = ds.apply_indices(first_asmt_indices, a_pids_)
            f_a_hlts_ = ds.apply_indices(first_asmt_indices, a_hlts_)
            print("  len(f_a_hlts_):", len(f_a_hlts_))
            d = dict()
            for i_r in range(len(f_a_hlts_)):
                d[f_a_pids_[i_r]] = f_a_hlts_[i_r]
            # print("  ", np.array_equal(ds.get_reader(flt_ptnts['id'])[:], f_a_pids_))
            p_ids_ = ds.get_reader(flt_ptnts['id'])[:]
            # print("  initially_health eq:", np.array_equal(f_a_pids_, p_ids_))
            p_init_hlt = np.zeros(len(p_ids_), dtype=np.bool)
            for i_p, p in enumerate(p_ids_):
                if p in d:
                    v = d[p]
                    p_init_hlt[i_p] = v
            # print("  check result eq:", np.array_equal(f_a_hlts_, p_init_hlt))
            ds.get_numeric_writer(flt_ptnts, 'initially_healthy', 'bool').write(p_init_hlt)

    # initial test filtering /flattening and date calculation
    # -------------------------------------------------------
    if 'filtered_tests' not in dest_data.keys():
        with utils.Timer("filtering tests", new_line=True):
            flt_tests = dest_data.create_group('filtered_tests')

            # Filter tests
            # ============

            t_cats = ds.get_reader(src_tests['created_at'])
            t_dts = ds.get_reader(src_tests['date_taken_specific'])
            t_dsbs = ds.get_reader(src_tests['date_taken_between_start'])
            t_dsbe = ds.get_reader(src_tests['date_taken_between_end'])
            t_effd_, t_filt_ =\
                effective_test_date(ds, start_timestamp, end_timestamp,
                                    t_cats, t_dts, t_dsbs, t_dsbe,
                                    flt_tests, 'eff_test_date', flt_tests, 'eff_test_filter')

            for k in ('patient_id', 'result'):
                reader = ds.get_reader(src_tests[k])
                writer = reader.get_writer(flt_tests, k)
                writer.write(ds.apply_filter(t_filt_, reader))
            etd_r = ds.get_reader(flt_tests['eff_test_date'])
            etd_w = etd_r.get_writer(flt_tests, 'eff_test_date', write_mode='overwrite')
            filt_effd_ = ds.apply_filter(t_filt_, t_effd_)
            etd_w.write(filt_effd_)

            # get test months
            t_month = ds.get_fixed_string_writer(flt_tests, 'test_month', 7)
            months_ = np.zeros(len(filt_effd_), dtype='S7')
            for i_r in range(len(months_)):
                dt = datetime.fromtimestamp(filt_effd_[i_r])
                months_[i_r] = '{}-{:02}'.format(dt.year, dt.month).encode()
            t_month.write(months_)
    else:
        print('filtered test fields already calculated')
        flt_tests = dest_data['filtered_tests']

    # join age, gender to assessments
    # -------------------------------
    if 'age' not in flt_asmts:
        with utils.Timer("mapping age / gender from filtered patients to filtered assessments",
                         new_line=True):
            p_ids_ = ds.get_reader(flt_ptnts['id'])[:]
            a_pids_ = ds.get_reader(flt_asmts['patient_id'])[:]
            with utils.Timer('  map assessment patient_ids to patient_ids'):
                asmt_to_ptnt = ds.get_index(p_ids_, a_pids_)

            with utils.Timer('  map patient age to assessment space'):
                ptnt_ages = ds.get_reader(flt_ptnts['age'])
                ptnt_ages.get_writer(flt_asmts, 'age').write(ptnt_ages[:][asmt_to_ptnt])

            with utils.Timer('  map patient gender to assessment gender'):
                ptnt_genders = ds.get_reader(flt_ptnts['gender'])
                ptnt_genders.get_writer(flt_asmts, 'gender').write(
                    ptnt_genders[:][asmt_to_ptnt])

            a_to_p = ds.get_index(p_ids_, a_pids_)

            a_ages_ = ds.apply_indices(a_to_p, ptnt_ages[:])
            ptnt_ages.get_writer(flt_asmts, 'age', write_mode='overwrite').write(a_ages_)
            a_genders_ = ds.apply_indices(a_to_p, ptnt_genders[:])
            ptnt_genders.get_writer(flt_asmts, 'gender',
                                    write_mode='overwrite').write(a_genders_)

    # calculate and select maximum predicted covid
    # --------------------------------------------
    if 'final_assessments' not in dest_data.keys():
        with utils.Timer('running covid prediction', new_line=True):
            t0 = time.time()
            scores = nature_medicine_model_1(ds.get_reader(flt_asmts['persistent_cough']),
                                             ds.get_reader(flt_asmts['skipped_meals']),
                                             ds.get_reader(flt_asmts['loss_of_smell']),
                                             ds.get_reader(flt_asmts['fatigue']),
                                             ds.get_reader(flt_asmts['age']),
                                             ds.get_reader(flt_asmts['gender']))
            ds.get_numeric_writer(flt_asmts, 'prediction_score', 'float32',
                                  writemode='overwrite').write(scores)

            predictions = scores > 0.5
            print("  covid prediction run in {}s".format(time.time() - t0))
            print("  covid prediction covid / not covid",
                  np.count_nonzero(predictions == 1),
                  np.count_nonzero(predictions == 0))
            ds.get_numeric_writer(flt_asmts, 'prediction', 'uint8',
                                  writemode='overwrite').write(predictions)

            # get the maximum prediction
            with utils.Timer("  getting spans for assessment patient_ids"):
                spans = ds.get_spans(ds.get_reader(flt_asmts['patient_id']))

            with utils.Timer("  getting max predictions for each patient"):
                max_score_indices = ds.apply_spans_index_of_max(spans, scores)

            print('  len(flt_ptnts["id"]:', len(ds.get_reader(flt_ptnts['id'])))
            print('  len(max_score_indices):', len(max_score_indices), max_score_indices[:10])

            fnl_asmts = dest_data.create_group('final_assessments')
            # apply the indices to the filtered assessments
            for k in flt_asmts.keys():
                with utils.Timer(f"applying selected_assessment indices to {k} "):
                    reader = ds.get_reader(flt_asmts[k])
                    ds.apply_indices(max_score_indices, reader,
                                   reader.get_writer(fnl_asmts, k))
                    # reader.get_writer(fnl_asmts, k).write(reader[:][max_score_indices])

            print(flt_asmts.keys())
            print(fnl_asmts.keys())

            flt_ptnts_ = ds.get_reader(flt_ptnts['id'])[:]
            fnl_asmts_ = ds.get_reader(fnl_asmts['patient_id'])[:]
            print('  len(flt_ptnts_):', len(flt_ptnts_))
            print('  len(fnl_asmts_):', len(fnl_asmts_))
            final_p_indices = ds.get_index(flt_ptnts_, fnl_asmts_)
            print("  len(final_p_indices):", len(final_p_indices))

            if not np.array_equal(flt_ptnts_[final_p_indices], fnl_asmts_):
                raise ValueError("filtered patients should map exactly with final assessments "
                                 "but don't!")

            gfpi = ds.get_or_create_group(dest_data, 'final_patient_index')
            ds.get_numeric_writer(gfpi, 'final_p_indices', 'int32',
                                  writemode='overwrite').write(final_p_indices)
            print(final_p_indices[:50])
    else:
        print('final_assessments already calculated')
        fnl_asmts = dest_data['final_assessments']

    # select maximum assessment test result and month
    # -----------------------------------------------
    if 'old_test_result' not in flt_ptnts.keys():
        with utils.Timer('collating max assessment test results / months', new_line=True):
            p_ids_ = ds.get_reader(flt_ptnts['id'])[:]
            print("  p_ids_", len(p_ids_))
            a_pids = ds.get_reader(flt_asmts['patient_id'])
            a_pids_ = a_pids[:]
            print("  a_pids_", len(a_pids_))
            a_cats = ds.get_reader(flt_asmts['created_at'])
            a_cats_ = a_cats[:]
            print("  a_cats_", len(a_cats_))
            a_tcps = ds.get_reader(flt_asmts['tested_covid_positive'])
            a_tcps_ = a_tcps[:]
            print("  a_tcps_", len(a_tcps_))
            asmt_spans = ds.get_spans(a_pids)

            asmt_test_indices = ds.apply_spans_index_of_max(asmt_spans, a_tcps)
            print('  len(asmt_test_indices):', len(asmt_test_indices))
            d = dict()
            for ia, a in enumerate(asmt_test_indices):
                d[a_pids_[a]] = (a_tcps_[a], a_cats_[a], ia)

            max_results_ = np.zeros(len(p_ids_), dtype=np.int8)
            result_months_ = np.zeros(len(p_ids_), dtype='S7')
            with utils.Timer("  mapping maximum assessment test result / month to patients"):
                for ip, p in enumerate(p_ids_):
                    if p in d:
                        v = d[p]
                        max_results_[ip] = v[0]
                        dt = datetime.fromtimestamp(v[1])
                        result_months_[ip] = '{}-{:02}'.format(dt.year, dt.month).encode()

            print(np.count_nonzero(max_results_ > 0))

            ds.get_numeric_writer(flt_ptnts, 'old_test_result', 'int8',
                                  writemode='overwrite').write(max_results_)
            ds.get_fixed_string_writer(flt_ptnts, 'old_test_month', 7,
                                       writemode='overwrite').write(result_months_)

    # select maximum test result and month
    # ------------------------------------
    if 'new_test_result' not in flt_ptnts.keys():
        with utils.Timer("calculating max test results and month", new_line=True):
            print("  ", flt_ptnts.keys())
            t_pids_ = ds.get_reader(flt_tests['patient_id'])[:]
            t_rslts_ = ds.get_reader(flt_tests['result'])[:]
            t_months_ = ds.get_reader(flt_tests['test_month'])[:]
            test_spans = ds.get_spans(t_pids_)
            max_result_indices = ds.apply_spans_index_of_max(test_spans, t_rslts_)

            d = dict()
            for i_t, t in enumerate(max_result_indices):
                d[t_pids_[t]] = (t_rslts_[t], t_months_[t], i_t)

            # build a dictionary for max results by pid
            p_ids_ = ds.get_reader(flt_ptnts['id'])[:]
            max_results_ = np.zeros(len(p_ids_), dtype=np.int8)
            result_months_ = np.zeros(len(p_ids_), dtype='S7')
            for i_p, p in enumerate(p_ids_):
                if p in d:
                    v = d[p]
                    max_results_[i_p] = v[0]
                    result_months_[i_p] = v[1]

            ds.get_numeric_writer(flt_ptnts, 'new_test_result', 'int8',
                                  writemode='overwrite').write(max_results_)
            ds.get_fixed_string_writer(flt_ptnts, 'new_test_month', 7,
                                       writemode='overwrite').write(result_months_)

    # verify that flt_patients and fnl_asmts are compatible
    # -----------------------------------------------------
    print('patients')
    print(flt_ptnts.keys())
    p_ids_ = ds.get_reader(flt_ptnts['id'])[:]
    print(len(p_ids_))

    print('assessments')
    print(fnl_asmts.keys())
    a_pids_ = ds.get_reader(fnl_asmts['patient_id'])[:]
    print(len(a_pids_))

    print(np.array_equal(p_ids_, a_pids_))

    initial_fields = ['id', '']
    asmt_fields = ['fatigue', 'loss_of_smell', 'persistent_cough', 'skipped_meals',
                   'hospitalization', 'treatment', 'prediction_score', 'prediction']

    # first_patient_fields = ['id', 'created_at', 'updated_at', 'version', 'country_code']
    # last_patient_fields = ['old_test_result', 'old_test_month', 'new_test_result', 'new_test_month']
    # patient_fields = [p for p in flt_ptnts.keys() if p not in first_patient_fields + last_patient_fields]
    # patient_fields = first_patient_fields + patient_fields + last_patient_fields
    # fields_to_export = [(p, flt_ptnts[p]) for p in patient_fields] + \
    #                    [('asmt_{}'.format(a), fnl_asmts[a]) for a in asmt_fields]
    # exporter.export_to_csv('/home/ben/covid/supplements_patients.csv', ds, fields_to_export)


if __name__ == '__main__':
    datastore = DataStore()
    src_file = '/home/ben/covid/ds_20200830_full.hdf5'
    dest_file = '/home/ben/covid/ds_20200830_supplements.hdf5'
    with h5py.File(src_file, 'r+') as src_data:
        with h5py.File(dest_file, 'w') as dest_data:
            supplement_paper_processing(datastore, src_data, dest_data)



import numpy as np

"""
Filter tests for well-formedness and sensible date range
Returns a filtered set of effective test 
"""
def effective_test_date(datastore, start_timestamp, end_timestamp,
                        created_at, date_taken_specific,
                        date_taken_between_start, date_taken_between_end,
                        dest_group, dest_name, filter_group, filter_name):

    raw_t_cats = created_at[:]
    raw_t_dts = date_taken_specific[:]
    raw_t_dsbs = date_taken_between_start[:]
    raw_t_dsbe = date_taken_between_end[:]
    test_filter = np.zeros(len(raw_t_cats), dtype=np.bool)

    # remove tests where no dates are set
    cur_filter = np.logical_not((raw_t_dts == 0) & (raw_t_dsbs == 0) & (raw_t_dsbe == 0))
    test_filter = cur_filter[:]
    print("standard test filter 1:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where all three dates are set
    cur_filter = np.logical_not((raw_t_dts != 0) & (raw_t_dsbs != 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 2:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where only one of the date range tests is set
    cur_filter = np.logical_not((raw_t_dsbs != 0) & (raw_t_dsbe == 0) |
                                (raw_t_dsbs == 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 3:", np.count_nonzero(test_filter), len(test_filter))

    eff_test_dates = np.where(test_filter == False,
                              np.nan,
                              np.where(raw_t_dts != 0,
                                       raw_t_dts,
                                       raw_t_dsbs + (raw_t_dsbe - raw_t_dsbs) / 2))

    # remove tests where the test date is after the created at date
    cur_filter = eff_test_dates <= raw_t_cats
    test_filter = test_filter & cur_filter
    print("standard test filter 7:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests with an effective test date outside permissible range
    cur_filter = (eff_test_dates >= start_timestamp) & (eff_test_dates <= end_timestamp)
    test_filter = test_filter & cur_filter
    print("standard_test_filter 8:", np.count_nonzero(test_filter), len(test_filter))

    print(len(eff_test_dates))
    print(len(test_filter))

    datastore.get_timestamp_writer(dest_group, dest_name,
                                   writemode='overwrite').write(eff_test_dates)
    datastore.get_numeric_writer(filter_group, filter_name, 'bool',
                                 writemode='overwrite').write(test_filter)

    return eff_test_dates, test_filter



import numpy as np

def consolidate_test_dates_v1(ds, created_at, date_taken_specific,
                              date_taken_between_start, date_taken_between_end,
                              effective_test_date,
                              initial_filter=None, first_timestamp=None, last_timestamp=None):
    raw_t_cats = created_at[:]
    raw_t_dts = date_taken_specific[:]
    raw_t_dsbs = date_taken_between_start[:]
    raw_t_dsbe = date_taken_between_end[:]

    if initial_filter:
        test_filter = initial_filter[:]

    # remove tests where no dates are set
    cur_filter = np.logical_not((raw_t_dts == 0) & (raw_t_dsbs == 0) & (raw_t_dsbe == 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 1:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where all three dates are set
    cur_filter = np.logical_not((raw_t_dts != 0) & (raw_t_dsbs != 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 2:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where only one of the date range tests is set
    cur_filter = np.logical_not((raw_t_dsbs != 0) & (raw_t_dsbe == 0) |
                                (raw_t_dsbs == 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 3:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where specific date is set but out of range
    cur_filter = \
        (raw_t_dts == 0) | ((raw_t_dts >= first_timestamp) & (raw_t_dts <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 4:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where beginning date is set but out of range
    cur_filter = \
        (raw_t_dsbs == 0) | ((raw_t_dsbs >= first_timestamp) & (raw_t_dsbs <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 5:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where ending date is set but out of range
    cur_filter = \
        (raw_t_dsbe == 0) | ((raw_t_dsbe >= first_timestamp) & (raw_t_dsbe <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 6:", np.count_nonzero(test_filter), len(test_filter))

    test_timestamps = np.where(raw_t_dts != 0,
                               raw_t_dts,
                               raw_t_dsbs + (raw_t_dsbe - raw_t_dsbs) / 2)

    # remove tests where the test date is after the created at date
    cur_filter = test_timestamps <= raw_t_cats
    test_filter = test_filter & cur_filter
    print("standard test filter 7:", np.count_nonzero(test_filter), len(test_filter))

    # t_rsts = ds.get_reader(s_tests['result'])
    # t_rsts.get_writer(flat_tests, 'result').write(ds.apply_filter(test_filter, t_rsts))
    # t_pids = ds.get_reader(s_tests['patient_id'])
    # t_pids.get_writer(flat_tests, 'patient_id').write(ds.apply_filter(test_filter, t_pids))
    effective_test_date.write(ds.apply_filter(test_filter, test_timestamps))



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import defaultdict
import csv
from datetime import datetime
from io import StringIO

import numpy as np
from numba import njit


def validate_file_exists(file_name):
    import os
    if not os.path.exists(file_name):
        raise FileExistsError(f"{file_name} doesn't exist'")
    if not os.path.isfile(file_name):
        raise FileNotFoundError(f"{file_name} is not a file'")


def find_longest_sequence_of(string, char):
    longest = 0
    current = 0
    for s in string:
        if s == char:
            current += 1
        else:
            if current > longest:
                longest = current
            current = 0
    if current > longest:
        longest = current
    return longest


def count_flag_empty(flags):
    count = 0
    for f in flags:
        if f == 0:
            count += 1
    return count


def count_flag_not_set(flags, flag_to_test):
    count = 0
    for f in flags:
        if not (f & flag_to_test):
            count += 1
    return count


def count_flag_set(flags, flag_to_test):
    count = 0
    for f in flags:
        if f & flag_to_test:
            count += 1
    return count


def timestamp_to_day(field):
    if field == '':
        return ''
    return f'{field[0:4]}-{field[5:7]}-{field[8:10]}'

def timestamp_to_datetime(field):
    try:
        ts = datetime.strptime(field, '%Y-%m-%d %H:%M:%S.%f%z')
    except ValueError:
        try:
            ts = datetime.strptime(field, '%Y-%m-%d %H:%M:%S%z')
        except ValueError:
            ts = datetime.strptime(field, '%y-%m-%d')

    return ts


def build_histogram(dataset, filtered_records=None, tx=None):
    # TODO: memory_efficiency: see build_histogram function
    histogram = defaultdict(int)
    for ir, r in enumerate(dataset):
        if not filtered_records or not filtered_records[ir]:
            if tx is not None:
                value = tx(r)
            else:
                value = r
            histogram[value] += 1
    hlist = list(histogram.items())
    del histogram
    return hlist


def filter_field(fields, filter_list, f_missing, f_bad, is_type_fn, type_fn, valid_fn):
    for ir, r in enumerate(fields):
        if not is_type_fn(r):
            if f_missing != 0:
                filter_list[ir] |= f_missing
        else:
            value = type_fn(r)
            if not valid_fn(value):
                if f_bad != 0:
                    filter_list[ir] |= f_bad


def map_between_categories(first_map, second_map):
    result_map = dict()
    for m in first_map.keys():
        result_map[first_map[m]] = second_map[m]
    return result_map


def to_categorical(field, transform):
    results = np.zeros_like(field, dtype=field.dtype)
    for ir, r in enumerate(field):
        results[ir] = transform[r]
    return results


def print_diagnostic_row(preamble, ds, ir, keys, fns=None):
    if fns is None:
        fns = dict()
    # indices = [ds.field_to_index(k) for k in keys]
    # indexed_fns = [None if k not in fns else fns[k] for k in keys]
    values = [None] * len(keys)
    # for ii, i in enumerate(indices):
    for i, k in enumerate(keys):
        if not fns or k not in fns:
            values[i] = ds.value_from_fieldname(ir, k)
        else:
            # values[i] = fns[k](ds.value_from_fieldname(ir, k))
            values[i] = ds.value_from_fieldname(ir, k)
        # if indexed_fns[ii] is None:
        #     values[ii] = fields[ir][i]
        # else:
        #     values[ii] = indexed_fns[ii](fields[ir][i])
    print(f'{preamble}: {values}')


def check_input_lengths(names, fields):
    assert(len(names) == len(fields))
    assert(isinstance(names, tuple))
    assert(isinstance(fields, tuple))

    length = None
    error = False
    for f in fields:
        if not length:
            length = len(f)
        elif length != len(f):
            error = True

    if error:
        field_name_str = ','.join([f"'{n}'" for n in names])
        length_str = ','.join([len(f) for f in fields])
        raise ValueError(f"Collections {field_name_str} have inconsistent lengths: ({length_str})")


def valid_range_fac(f_min, f_max, default_value=''):
    def inner_(x):
        return x == default_value or x > f_min and x < f_max
    return inner_


def valid_range_fac_inc(f_min, f_max, default_value=''):
    def inner_(x):
        return x == default_value or x >= f_min and x <= f_max
    return inner_


def datetime_to_seconds(dt):
    return f'{dt[0:4]}-{dt[5:7]}-{dt[8:10]} {dt[11:13]}:{dt[14:16]}:{dt[17:19]}'


def is_int(value):
    try:
        int(float(value))
        return True
    except:
        return False


def is_float(value):
    try:
        float(value)
        return True
    except:
        return False


def to_int(value):
    try:
        fvalue = float(value)
    except ValueError as e:
        raise ValueError(f'{value} cannot be converted to float')

    try:
        ivalue = int(fvalue)
    except ValueError as e:
        raise ValueError(f'{fvalue} cannot be converted to int')

    return ivalue


def to_float(value):
    try:
        fvalue = float(value)
    except ValueError as e:
        raise ValueError(f'{value} cannot be converted to float')

    return fvalue


def list_to_escaped(strings):
    s = StringIO()
    w = csv.writer(s)
    w.writerow(strings)
    return s.getvalue()


def to_escaped(string, separator=',', delimiter='"'):
    comma = False
    quotes = False
    for c in string:
        if c == separator:
            comma = True
        elif c == delimiter:
            quotes = True

    if comma or quotes:
        s = StringIO()
        s.write(delimiter)
        for c in string:
            if c == delimiter:
                s.write(c)
            s.write(c)
        s.write(delimiter)
        return s.getvalue()
    else:
        return string

@njit
def bytearray_to_escaped(srcbytearray, destbytearray,
                         src_start=np.int64(0), src_end=None, dest_start=np.int64(0),
                         separator=b',', delimiter=b'"'):
    if src_end is None:
        src_end = np.int64(len(srcbytearray))

    comma = False
    quotes = False
    for i_c in range(src_start, src_end):
        c = srcbytearray[i_c]
        if c == separator:
            comma = True
        elif c == delimiter:
            quotes = True

    d_index = 0
    if comma or quotes:
        destbytearray[d_index] = delimiter
        d_index += 1
        for i_c in range(src_start, src_end):
            c = srcbytearray[i_c]
            if c == delimiter:
                destbytearray[d_index] = c
                d_index += 1
            destbytearray[d_index] = c
            d_index += 1
        destbytearray[d_index] = delimiter
        d_index += 1
        return d_index
    else:
        s_len = np.int64(src_end - src_start)
        destbytearray[dest_start:dest_start + s_len] = srcbytearray[src_start:src_end]
        d_index += s_len
        return d_index


def from_escaped(string):
    s = StringIO(string)
    r = csv.reader(s)
    return next(r)


def concatenate_maybe_strs(sequence, value, separator=',', delimiter='"'):
    if sequence in (None, '', 'NA'):
        if value in (None, '', 'NA'):
            return ''
        else:
            return to_escaped(value)
    else:
        if value in (None, '', 'NA'):
            return sequence
        else:
            return f"{sequence}{separator}{to_escaped(value)}"


def replace_if_invalid(replacement):
    def inner_(value):
        if value is '':
            return replacement
        else:
            return float(value)
    return inner_


def clear_set_flag(values, to_clear):
    for v in range(len(values)):
        values[v] &= ~to_clear
    return values


def sort_mixed_list(values, check_fn, sort_fn):
    # pass to find the single entry that fails check_fn
    for iv in range(len(values)):
        checked_item = None
        if not check_fn(values[iv]):
            #swap the current item with the last if it isn't last
            found_checked_item = True
            if iv != len(values) - 1:
                values[iv], values[-1] = values[-1], values[iv]
                checked_item = values.pop()
        break

    list.sort(values, key=sort_fn)
    if found_checked_item:
        values.append(checked_item)

    return values


class Timer:
    def __init__(self, start_msg, new_line=False, end_msg='completed in'):
        print(start_msg, end=': ' if new_line is False else '\n')
        self.end_msg = end_msg

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(self.end_msg + f' {time.time() - self.t0} seconds')



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hystore.core import dataset, utils

# fn1 = '/home/ben/covid/patients_export_geocodes_20200406050002.csv'
# fn2 = '/home/ben/covid/patients_export_geocodes_20200413050002.csv'

fn1 = '/home/ben/covid/patients_export_geocodes_20200413050002.csv'
fn2 = '/home/ben/covid/patients_export_geocodes_20200416050002.csv'

print('loading file 1')
with open(fn1) as f1:
    ds1 = dataset.Dataset(f1)
ds1.sort(('id',))
print('done')

print(); print('loading_file 2')
with open(fn2) as f2:
    ds2 = dataset.Dataset(f2)
ds2.sort(('id',))
print('done')

fields = ('id', 'updated_at', 'year_of_birth', 'height_cm', 'zipcode', 'outward_postcode')
i = 0
j = 0
matches = 0
mismatches = 0
empty = 0
zip1 = ds1.field_to_index('zipcode')
zip2 = ds2.field_to_index('zipcode')
owp1 = ds1.field_to_index('outward_postcode')
owp2 = ds2.field_to_index('outward_postcode')
while i < ds1.row_count() and j < ds2.row_count():
    if ds1.value(i, 0) < ds2.value(j, 0):
        i += 1
    elif ds1.value(i, 0) > ds2.value(j, 0):
        j += 1
    else:
        if i < 100:
            utils.print_diagnostic_row(f'{matches} {i}:', ds1, ds1.fields_, i, fields)
            utils.print_diagnostic_row(f'{matches} {j}:', ds2, ds2.fields_, j, fields)
        vzip1 = ds1.value(i, zip1)
        vzip2 = ds2.value(j, zip2)
        vowp1 = ds1.value(i, owp1)
        vowp2 = ds2.value(j, owp2)
        if vzip1 == '' and vzip2 == '' and vowp1 == '' and vowp2 == '':
            empty += 1
        elif vzip1 == vzip2 and vowp1 == vowp2:
            matches += 1
        else:
            mismatches += 1

        if i > 0 and i % 100000 == 0:
            print(f'{i} {matches}, {mismatches}, {empty}')
        i += 1
        j += 1

print('empty:', empty)
print('matches:', matches)
print('mismatches:', mismatches)


# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class CalculateMergedFieldCount:
    def __init__(self, updated_ats):
        self.updated_ats = updated_ats
        self.merged_row_count = 0

    def __call__(self, patient_id, dummy_filter_status, start, end):
        for i in range(start + 1, end + 1):
            last_date_str = self.updated_ats[i - 1]
            last_date = (last_date_str[0:4], last_date_str[5:7], last_date_str[8:10])
            cur_date_str = self.updated_ats[i]
            cur_date = (cur_date_str[0:4], cur_date_str[5:7], cur_date_str[8:10])
            if last_date == cur_date:
                self.merged_row_count += 1


class MergeAssessmentRows:
    def __init__(self, concat_field_indices,
                 resulting_fields, created_fields, existing_field_indices,
                 custom_field_aggregators,
                 source_filter, resulting_filter):
        print(created_fields.keys())
        self.rfindex = 0
        self.concat_indices = concat_field_indices
        self.source_filter = source_filter
        self.resulting_filter = resulting_filter
        self.resulting_fields = resulting_fields
        self.created_fields = created_fields
        self.existing_field_indices = existing_field_indices
        self.custom_field_aggregators = custom_field_aggregators


    def populate_row(self, source_fields, source_index):
        for e in self.existing_field_indices:
            if e[0] in self.custom_field_aggregators:
                self.resulting_fields[e[0]][self.rfindex] =\
                    self.custom_field_aggregators[e[0]](
                        self.resulting_fields[e[0]][self.rfindex],
                        source_fields[e[1]][source_index])
            else:
                self.resulting_fields[e[0]][self.rfindex] = source_fields[e[1]][source_index]
        for ck, cv in self.created_fields.items():
            self.resulting_fields[ck][self.rfindex] =\
                max(self.resulting_fields[ck][self.rfindex], cv[source_index])
        self.resulting_filter[self.rfindex] |= self.source_filter[source_index]

    def __call__(self, fields, dummy, start, end):
        # first pass: determine escape sequences for fields that need concatenation
        # esq_sequences = [1] * len(self.concat_indices)
        # for i in range(start + 1, end + 1):
        #     for i_c, c in enumerate(self.concat_indices):
        #         esq_sequences[i_c] =\
        #             max(esq_sequences[i_c], find_longest_sequence_of(fields[c][i], '`'))
        # if esq_sequences != [1, 1]:
        #     print(fields[1], esq_sequences)
        # write the first row to the current resulting field index
        prev_date_str = fields[3][start]
        prev_date = (prev_date_str[0:4], prev_date_str[5:7], prev_date_str[8:10])
        self.populate_row(fields, start)

        for i in range(start + 1, end + 1):
            cur_date_str = fields[3][i]
            cur_date = (cur_date_str[0:4], cur_date_str[5:7], cur_date_str[8:10])
            if cur_date != prev_date:
                self.rfindex += 1
            if i % 1000000 == 0 and i > 0:
                print('.')
            self.populate_row(fields, i)
            prev_date = cur_date

        # finally, update the resulting field index one more time
        self.rfindex += 1



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hystore.core import utils


def na_or_value(value):
    if value == '':
        return 'na'
    return value

def na_compare(value1, value2):
    lv1 = value1.lower()
    lv2 = value2.lower()
    if lv1 == 'na' and lv2 in ('', 'na'):
        return True
    return lv1 == lv2

def datetime_compare_to_secs(value1, value2):
    dt1 = utils.datetime_to_seconds(value1)
    dt2 = utils.datetime_to_seconds(value2)
    return dt1 == dt2

def check_row(exp_ds, exp_index, act_ds, act_index, keys, custom_checks):
    disparities = None
    for k in keys:
        if isinstance(k, str):
            kexp, kact = k, k
        else:
            kexp, kact = k
        if kexp in custom_checks:
            check = custom_checks[kexp]
        else:
            check = na_compare
        exp_value = exp_ds.field_by_name(kexp)[exp_index]
        act_value = act_ds.field_by_name(kact)[act_index]
        if not check(exp_value, act_value):
            if disparities is None:
                disparities = list()
            if kexp == kact:
                disparities.append(f'{kexp} : {str(na_or_value(exp_value))} / {str(na_or_value(act_value))}')
            else:
                disparities.append(f'{kexp} / {kact} : {str(na_or_value(exp_value))} / {str(na_or_value(act_value))}')
    return disparities


# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class ValidateCovidTestResultsFacVersion1PreHCTFix:
    def __init__(self, hcts, tcps, filter_status, hct_results, results, filter_flag, show_debug=False):
        self.valid_transitions = {0: (0, 1, 2, 3), 1: (0, 1, 2, 3), 2: (0, 2), 3: (0, 3)}
        self.upgrades = {0: (0, 1, 2, 3), 1: (2, 3), 2: tuple(), 3: tuple()}
        self.hcts = hcts
        self.tcps = tcps
        self.hct_results = hct_results
        self.results = results
        self.filter_status = filter_status
        self.filter_flag = filter_flag
        self.show_debug = show_debug

    def __call__(self, patient_id, filter_status, start, end):
        # validate the subrange
        invalid = False
        max_value = 0
        for j in range(start, end + 1):
            # allowable transitions
            value = self.tcps[j]
            if value not in self.valid_transitions[max_value]:
                invalid = True
                break
            if value in self.upgrades[max_value]:
                max_value = value
            self.results[j] = max_value

        if invalid:
            for j in range(start, end + 1):
                self.results[j] = self.tcps[j]
                filter_status[j] |= self.filter_flag

        self.hct_results[start:end+1] = self.hcts[start:end+1]
        if invalid:
            for j in range(start, end + 1):
                filter_status[j] |= self.filter_flag

        if self.show_debug == True:
            if invalid or not np.array_equal(self.hcts[start:end+1], self.hct_results[start:end+1])\
               or not np.array_equal(self.tcps[start:end+1], self.results[start:end+1]):
                reason = 'inv' if invalid else 'diff'
                print(reason, start, 'hct:', self.hcts[start:end+1], self.hct_results[start:end+1])
                print(reason, start, 'tcp:', self.tcps[start:end+1], self.results[start:end+1])

        # TODO: remove before v0.1.8
        # for j in range(start, end + 1):
        #     self.hct_results[j] = self.hcts[j]

        # if invalid:
        #     for j in range(start, end + 1):
        #         if self.hct_results[j] == 1 and self.results[j] != 0:
        #             print('hct:', start, self.hcts[start:end+1], self.hct_results[start:end+1])
        #             print('tcp:', start, self.tcps[start:end+1], self.results[start:end+1])
        #             break


class ValidateCovidTestResultsFacVersion1:
    def __init__(self, hcts, tcps, filter_status, hct_results, results, filter_flag, show_debug=False):
        self.valid_transitions = {0: (0, 1, 2, 3), 1: (0, 1, 2, 3), 2: (0, 2), 3: (0, 3)}
        self.upgrades = {0: (0, 1, 2, 3), 1: (2, 3), 2: tuple(), 3: tuple()}
        self.hcts = hcts
        self.tcps = tcps
        self.hct_results = hct_results
        self.results = results
        self.filter_status = filter_status
        self.filter_flag = filter_flag
        self.show_debug = show_debug

    def __call__(self, patient_id, filter_status, start, end):
        # validate the subrange
        invalid = False
        max_value = 0
        for j in range(start, end + 1):
            # allowable transitions
            value = self.tcps[j]
            if value not in self.valid_transitions[max_value]:
                invalid = True
                break
            if value in self.upgrades[max_value]:
                max_value = value
            self.results[j] = max_value

        if invalid:
            for j in range(start, end + 1):
                self.results[j] = self.tcps[j]
                filter_status[j] |= self.filter_flag

        if not invalid:
            first_hct_false = -1
            first_hct_true = -1
            for j in range(start, end + 1):
                if self.hcts[j] == 1:
                    if first_hct_false == -1:
                        first_hct_false = j
                elif self.hcts[j] == 2:
                    if first_hct_true == -1:
                        first_hct_true = j

            max_value = 0
            for j in range(start, end + 1):
                if j == first_hct_false:
                    max_value = max(max_value, 1)
                if j == first_hct_true:
                    max_value = 2

                if self.hct_results is None or self.hct_results[j] is None:
                    print(j, self.hct_results[j], max_value)
                self.hct_results[j] = max_value
        else:
            for j in range(start, end + 1):
                self.hct_results[j] = self.hcts[j]
                filter_status[j] |= self.filter_flag

        if self.show_debug == True:
            if invalid or not np.array_equal(self.hcts[start:end+1], self.hct_results[start:end+1])\
               or not np.array_equal(self.tcps[start:end+1], self.results[start:end+1]):
                reason = 'inv' if invalid else 'diff'
                print(reason, start, 'hct:', self.hcts[start:end+1], self.hct_results[start:end+1])
                print(reason, start, 'tcp:', self.tcps[start:end+1], self.results[start:end+1])

        # TODO: remove before v0.1.8
        # for j in range(start, end + 1):
        #     self.hct_results[j] = self.hcts[j]

        # if invalid:
        #     for j in range(start, end + 1):
        #         if self.hct_results[j] == 1 and self.results[j] != 0:
        #             print('hct:', start, self.hcts[start:end+1], self.hct_results[start:end+1])
        #             print('tcp:', start, self.tcps[start:end+1], self.results[start:end+1])
        #             break


class ValidateCovidTestResultsFacVersion2:
    def __init__(self, hcts, tcps, filter_status, hct_results, results, filter_flag, show_debug=False):
        self.valid_transitions = {0: (0, 1, 2, 3), 1: (0, 1, 2, 3), 2: (0, 2), 3: (0, 3)}
        self.valid_transitions_before_yes =\
            {0: (0, 1, 2, 3), 1: (0, 1, 2, 3), 2: (0, 1, 2, 3), 3: (0, 3)}
        self.upgrades = {0: (0, 1, 2, 3), 1: (2, 3), 2: tuple(), 3: tuple()}
        self.upgrades_before_yes = {0: (1, 2, 3), 1: (3,), 2: (1, 3), 3: tuple()}

        self.tcps = tcps
        self.results = results
        self.filter_status = filter_status
        self.filter_flag = filter_flag
        self.show_debug = show_debug

    def __call__(self, patient_id, filter_status, start, end):
        # validate the subrange
        invalid = False
        max_value = 0
        first_waiting = -1
        first_no = -1
        first_yes = -1

        for j in range(start, end + 1):
            value = self.tcps[j]
            if value == 'waiting' and first_waiting == -1:
                first_waiting = j
            if value == 'no' and first_no == -1:
                first_no = j
            if value == 'yes' and first_yes == -1:
                first_yes = j

        for j in range(start, end + 1):
            valid_transitions = self.valid_transitions_before_yes if j <= first_yes else self.valid_transitions
            upgrades = self.upgrades_before_yes if j <= first_yes else self.upgrades
            # allowable transitions
            value = self.tcps[j]
            if value not in valid_transitions[max_value]:
                invalid = True
                break
            if j < first_yes and value == 'no':
                value == 'waiting'
            if value in upgrades[max_value]:
                max_value = value
            self.results[j] = max_value

        #rescue na -> waiting -> no -> waiting
        if invalid and first_yes == -1 and self.tcps[end] == 'waiting':
            invalid = False
            max_value = ''
            for j in range(start, end+1):
                value = self.tcps[j]
                if max_value == '' and value != '':
                    max_value = 'waiting'
                self.results[j] = max_value

        if invalid:
            for j in range(start, end + 1):
                self.results[j] = self.tcps[j]
                filter_status[j] |= self.filter_flag
            if self.show_debug:
                print(self.tcps[j], end=': ')
                for j in range(start, end + 1):
                    if j > start:
                        print(' ->', end=' ')
                    value = self.tcps[j]
                    print('na' if value == '' else value, end='')
                print('')


from datetime import datetime, timezone

import numpy as np
import numba
import h5py

from hystore.core.readerwriter import DataWriter
from hystore.core import utils


# def test_field_iterator(data):
#     @numba.njit
#     def _inner():
#         for d in data:
#             yield d
#     return _inner()
#
# iterator_type = numba.from_dtype(test_field_iterator)
#
# @numba.jit
# def sum_iterator(iter_):
#     total = np.int64(0)
#     for i in iter_:
#         total += i
#     return total


class Field:
    def __init__(self, session, group, name=None, write_enabled=False):
        # if name is None, the group is an existing field
        # if name is set but group[name] doesn't exist, then create the field
        if name is None:
            # the group is an existing field
            field = group
        else:
            field = group[name]
        self._session = session
        self._field = field
        self._fieldtype = self._field.attrs['fieldtype']
        self._write_enabled = write_enabled
        self._value_wrapper = None

    @property
    def name(self):
        return self._field.name

    @property
    def timestamp(self):
        return self._field.attrs['timestamp']

    @property
    def chunksize(self):
        return self._field.attrs['chunksize']


class ReadOnlyFieldArray:
    def __init__(self, field, dataset_name):
        self._field = field
        self._name = dataset_name
        self._dataset = field[dataset_name]

    def __len__(self):
        return len(self._dataset)

    @property
    def dtype(self):
        return self._dataset.dtype

    def __getitem__(self, item):
        return self._dataset[item]

    def __setitem__(self, key, value):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def clear(self):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def write_part(self, part):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def write(self, part):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def complete(self):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")


class WriteableFieldArray:
    def __init__(self, field, dataset_name):
        self._field = field
        self._name = dataset_name
        self._dataset = field[dataset_name]

    def __len__(self):
        return len(self._dataset)

    @property
    def dtype(self):
        return self._dataset.dtype

    def __getitem__(self, item):
        return self._dataset[item]

    def __setitem__(self, key, value):
        self._dataset[key] = value

    def clear(self):
        DataWriter._clear_dataset(self._field, self._name)

    def write_part(self, part):
        DataWriter.write(self._field, self._name, part, len(part), dtype=self._dataset.dtype)

    def write(self, part):
        DataWriter.write(self._field, self._name, part, len(part), dtype=self._dataset.dtype)
        self.complete()

    def complete(self):
        DataWriter.flush(self._field[self._name])


class ReadOnlyIndexedFieldArray:
    def __init__(self, field, index_name, values_name):
        self._field = field
        self._index_name = index_name
        self._index_dataset = field[index_name]
        self._values_name = values_name
        self._values_dataset = field[values_name]

    def __len__(self):
        return len(self._index_dataset)-1

    def __getitem__(self, item):
        try:
            if isinstance(item, slice):
                start = item.start if item.start is not None else 0
                stop = item.stop if item.stop is not None else len(self._index_dataset) - 1
                step = item.step
                #TODO: validate slice
                index = self._index_dataset[start:stop+1]
                bytestr = self._values_dataset[index[0]:index[-1]]
                results = [None] * (len(index)-1)
                startindex = start
                for ir in range(len(results)):
                    results[ir] =\
                        bytestr[index[ir]-np.int64(startindex):
                                index[ir+1]-np.int64(startindex)].tobytes().decode()
                return results
        except Exception as e:
            print("{}: unexpected exception {}".format(self._field.name, e))
            raise

    def __setitem__(self, key, value):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def clear(self):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def write_part(self, part):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def write(self, part):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")

    def complete(self):
        raise PermissionError("This field was created read-only; call <field>.writeable() "
                              "for a writeable copy of the field")


class WriteableIndexedFieldArray:
    def __init__(self, field, index_name, values_name):
        self._field = field
        self._index_name = index_name
        self._index_dataset = field[index_name]
        self._values_name = values_name
        self._values_dataset = field[values_name]
        self._chunksize = self._field.attrs['chunksize']
        self._raw_values = np.zeros(self._chunksize, dtype=np.uint8)
        self._raw_indices = np.zeros(self._chunksize, dtype=np.int64)
        self._accumulated = self._index_dataset[-1]
        self._index_index = 0
        self._value_index = 0

    def __len__(self):
        return len(self._index_dataset) - 1

    def __getitem__(self, item):
        try:
            if isinstance(item, slice):
                start = item.start if item.start is not None else 0
                stop = item.stop if item.stop is not None else len(self._index_dataset) - 1
                step = item.step
                # TODO: validate slice
                index = self._index_dataset[start:stop + 1]
                bytestr = self._values_dataset[index[0]:index[-1]]
                results = [None] * (len(index) - 1)
                startindex = start
                rmax = min(len(results), stop - start)
                for ir in range(rmax):
                    rbytes = bytestr[index[ir] - np.int64(startindex):
                                index[ir + 1] - np.int64(startindex)].tobytes()
                    rstr = rbytes.decode()
                    results[ir] = rstr
                return results
        except Exception as e:
            print("{}: unexpected exception {}".format(self._field.name, e))
            raise

    def __setitem__(self, key, value):
        raise PermissionError("IndexedStringField instances cannot be edited via array syntax;"
                              "use clear and then write/write_part or write_raw/write_part_raw")

    def clear(self):
        self._accumulated = 0
        DataWriter.clear_dataset(self._field, self._index_name)
        DataWriter.clear_dataset(self._field, self._values_name)
        DataWriter.write(self._field, self._index_name, [0], 1, 'int64')
        DataWriter.write(self._field, self._values_name, [], 0, 'uint8')
        self._index_dataset = self._field[self._index_name]
        self._values_dataset = self._field[self._values_name]
        self._accumulated = 0


    def write_part(self, part):
        for s in part:
            evalue = s.encode()
            for v in evalue:
                self._raw_values[self._value_index] = v
                self._value_index += 1
                if self._value_index == self._chunksize:
                    DataWriter.write(self._field, self._values_name,
                                     self._raw_values, self._value_index)
                    self._value_index = 0
                self._accumulated += 1
            self._raw_indices[self._index_index] = self._accumulated
            self._index_index += 1
            if self._index_index == self._chunksize:
                DataWriter.write(self._field, self._index_name,
                                 self._raw_indices, self._index_index)
                self._index_index = 0


    def write(self, part):
        self.write_part(part)
        self.complete()

    def complete(self):
        if self._value_index != 0:
            DataWriter.write(self._field, self._values_name,
                             self._raw_values, self._value_index)
            self._value_index = 0
        if self._index_index != 0:
            DataWriter.write(self._field, self._index_name,
                             self._raw_indices, self._index_index)
            self._index_index = 0


def base_field_contructor(session, group, name, timestamp=None, chunksize=None):
    if name in group:
        msg = "Field '{}' already exists in group '{}'"
        raise ValueError(msg.format(name, group))

    field = group.create_group(name)
    field.attrs['chunksize'] = session.chunksize if chunksize is None else chunksize
    field.attrs['timestamp'] = session.chunksize if chunksize is None else chunksize
    return field


def indexed_string_field_constructor(session, group, name, timestamp=None, chunksize=None):
    field = base_field_contructor(session, group, name, timestamp, chunksize)
    field.attrs['fieldtype'] = 'indexedstring'
    DataWriter.write(field, 'index', [0], 1, 'int64')
    DataWriter.write(field, 'values', [], 0, 'uint8')


def fixed_string_field_constructor(session, group, name, length, timestamp=None, chunksize=None):
    field = base_field_contructor(session, group, name, timestamp, chunksize)
    field.attrs['fieldtype'] = 'fixedstring,{}'.format(length)
    field.attrs['strlen'] = length
    DataWriter.write(field, 'values', [], 0, "S{}".format(length))


def numeric_field_constructor(session, group, name, nformat, timestamp=None, chunksize=None):
    field = base_field_contructor(session, group, name, timestamp, chunksize)
    field.attrs['fieldtype'] = 'numeric,{}'.format(nformat)
    field.attrs['nformat'] = nformat
    DataWriter.write(field, 'values', [], 0, nformat)


def categorical_field_constructor(session, group, name, nformat, key,
                                  timestamp=None, chunksize=None):
    field = base_field_contructor(session, group, name, timestamp, chunksize)
    field.attrs['fieldtype'] = 'categorical,{}'.format(nformat)
    field.attrs['nformat'] = nformat
    DataWriter.write(field, 'values', [], 0, nformat)
    key_values = [v for k, v in key.items()]
    key_names = [k for k, v in key.items()]
    DataWriter.write(field, 'key_values', key_values, len(key_values), 'int8')
    DataWriter.write(field, 'key_names', key_names, len(key_names), h5py.special_dtype(vlen=str))


def timestamp_field_constructor(session, group, name, timestamp=None, chunksize=None):
    field = base_field_contructor(session, group, name, timestamp, chunksize)
    field.attrs['fieldtype'] = 'timestamp'
    DataWriter.write(field, 'values', [], 0, 'float64')


class IndexedStringField(Field):
    def __init__(self, session, group, name=None, write_enabled=False):
        super().__init__(session, group, name=name, write_enabled=write_enabled)
        self._session = session
        self._data_wrapper = None
        self._index_wrapper = None
        self._value_wrapper = None

    def writeable(self):
        return IndexedStringField(self._session, self._field, write_enabled=True)

    def create_like(self, group, name, timestamp=None):
        ts = self.timestamp if timestamp is None else timestamp
        indexed_string_field_constructor(self._session, group, name, ts, self.chunksize)
        return IndexedStringField(self._session, group, name, write_enabled=True)

    @property
    def data(self):
        if self._data_wrapper is None:
            wrapper =\
                WriteableIndexedFieldArray if self._write_enabled else ReadOnlyIndexedFieldArray
            self._data_wrapper = wrapper(self._field, 'index', 'values')
        return self._data_wrapper

    @property
    def indices(self):
        if self._index_wrapper is None:
            wrapper = WriteableFieldArray if self._write_enabled else ReadOnlyFieldArray
            self._index_wrapper = wrapper(self._field, 'index')
        return self._index_wrapper

    @property
    def values(self):
        if self._value_wrapper is None:
            wrapper = WriteableFieldArray if self._write_enabled else ReadOnlyFieldArray
            self._value_wrapper = wrapper(self._field, 'values')
        return self._value_wrapper


class FixedStringField(Field):
    def __init__(self, session, group, name=None, write_enabled=False):
        super().__init__(session, group, name=name, write_enabled=write_enabled)

    def writeable(self):
        return FixedStringField(self._session, self._field, write_enabled=True)

    def create_like(self, group, name, timestamp=None):
        ts = self.timestamp if timestamp is None else timestamp
        length = self._field.attrs['strlen']
        fixed_string_field_constructor(self._session, group, name, length, ts, self.chunksize)
        return FixedStringField(self._session, group, name, write_enabled=True)

    @property
    def data(self):
        if self._value_wrapper is None:
            if self._write_enabled:
                self._value_wrapper = WriteableFieldArray(self._field, 'values')
            else:
                self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
        return self._value_wrapper


class NumericField(Field):
    def __init__(self, session, group, name=None, write_enabled=False):
        super().__init__(session, group, name=name, write_enabled=write_enabled)

    def writeable(self):
        return NumericField(self._session, self._field, write_enabled=True)

    def create_like(self, group, name, timestamp=None):
        ts = self.timestamp if timestamp is None else timestamp
        nformat = self._field.attrs['nformat']
        numeric_field_constructor(self._session, group, name, nformat, ts, self.chunksize)
        return NumericField(self._session, group, name, write_enabled=True)

    @property
    def data(self):
        if self._value_wrapper is None:
            if self._write_enabled:
                self._value_wrapper = WriteableFieldArray(self._field, 'values')
            else:
                self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
        return self._value_wrapper


class CategoricalField(Field):
    def __init__(self, session, group,
                 name=None, write_enabled=False):
        super().__init__(session, group, name=name, write_enabled=write_enabled)

    def writeable(self):
        return CategoricalField(self._session, self._field, write_enabled=True)

    def create_like(self, group, name, timestamp=None):
        ts = self.timestamp if timestamp is None else timestamp
        nformat = self._field.attrs['nformat'] if 'nformat' in self._field.attrs else 'int8'
        keys = {v: k for k, v in self.keys.items()}
        categorical_field_constructor(self._session, group, name, nformat, keys,
                                      ts, self.chunksize)
        return CategoricalField(self._session, group, name, write_enabled=True)

    @property
    def data(self):
        if self._value_wrapper is None:
            if self._write_enabled:
                self._value_wrapper = WriteableFieldArray(self._field, 'values')
            else:
                self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
        return self._value_wrapper


    # Note: key is presented as value: str, even though the dictionary must be presented
    # as str: value
    @property
    def keys(self):
        kv = self._field['key_values']
        kn = self._field['key_names']
        keys = dict(zip(kv, kn))
        return keys


class TimestampField(Field):
    def __init__(self, session, group, name=None, write_enabled=False):
        super().__init__(session, group, name=name, write_enabled=write_enabled)

    def writeable(self):
        return TimestampField(self._session, self._field, write_enabled=True)

    def create_like(self, group, name, timestamp=None):
        ts = self.timestamp if timestamp is None else timestamp
        timestamp_field_constructor(self._session, group, name, ts, self.chunksize)
        return TimestampField(self._session, group, name, write_enabled=True)

    @property
    def data(self):
        if self._value_wrapper is None:
            if self._write_enabled:
                self._value_wrapper = WriteableFieldArray(self._field, 'values')
            else:
                self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
        return self._value_wrapper


class IndexedStringImporter:
    def __init__(self, session, group, name, timestamp=None, chunksize=None):
        indexed_string_field_constructor(session, group, name, timestamp, chunksize)
        self._field = IndexedStringField(session, group, name, write_enabled=True)

    def chunk_factory(self, length):
        return [None] * length

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            self._field.data.write_part(values)

    def complete(self):
        self._field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class FixedStringImporter:
    def __init__(self, session, group, name, length, timestamp=None, chunksize=None):
        fixed_string_field_constructor(session, group, name, length, timestamp, chunksize)
        self._field = FixedStringField(session, group, name, write_enabled=True)

    def chunk_factory(self, length):
        return np.zeros(length, dtype=self._field.data.dtype)

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            # self._field.data.write_part([v.encode() for v in values])
            self._field.data.write_part(values)

    def complete(self):
        self._field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class NumericImporter:
    def __init__(self, session, group, name, dtype, parser, timestamp=None, chunksize=None):
        numeric_field_constructor(session, group, name, dtype, timestamp, chunksize)
        numeric_field_constructor(session, group, '{}_valid'.format(name), 'bool',
                                  timestamp, chunksize)

        chunksize = session.chunksize if chunksize is None else chunksize
        self._field = NumericField(session, group, name, write_enabled=True)
        self._filter_field = NumericField(session, group, name, write_enabled=True)

        self._parser = parser
        self._values = np.zeros(chunksize, dtype=self._field.data.dtype)

        self._filter_values = np.zeros(chunksize, dtype='bool')

    def chunk_factory(self, length):
        # return np.zeros(length, dtype=self._field.data.dtype)
        return [None] * length

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            for i in range(len(values)):
                valid, value = self._parser(values[i])
                self._values[i] = value
                self._filter_values[i] = valid
            self._field.data.write_part(self._values)
            self._filter_field.data.write_part(self._filter_values)

    def complete(self):
        self._field.data.complete()
        self._filter_field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class CategoricalImporter:
    def __init__(self, session, group, name, value_type, keys, timestamp=None, chunksize=None):
        chunksize = session.chunksize if chunksize is None else chunksize
        categorical_field_constructor(session, group, name, value_type, keys, timestamp, chunksize)
        self._field = CategoricalField(session, group, name, write_enabled=True)
        self._keys = keys
        self._dtype = value_type
        self._key_type = 'U{}'.format(max(len(k.encode()) for k in keys))
        # self._results = np.zeros(chunksize, dtype=value_type)

    def chunk_factory(self, length):
        # return np.zeros(length, dtype=self._key_type)
        return np.zeros(length, dtype=self._dtype)

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            keys = self._keys
            # results = self._results
            # for i in range(len(values)):
            #     results = keys[values[i]]
            # self._field.data.write_part(results)
            self._field.data.write_part(values)

    def complete(self):
        self._field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class LeakyCategoricalImporter:
    def __init__(self, session, group, name, value_type, keys, out_of_range,
                 timestamp=None, chunksize=None):
        chunksize = session.chunksize if chunksize is None else chunksize
        categorical_field_constructor(session, group, name, value_type, keys,
                                      timestamp, chunksize)
        out_of_range_name = '{}_{}'.format(name, out_of_range)
        indexed_string_field_constructor(session, group, out_of_range_name,
                                         timestamp, chunksize)

        self._field = CategoricalField(session, group, name, write_enabled=True)
        self._str_field = IndexedStringField(session, group, out_of_range_name, write_enabled=True)

        self._keys = keys
        self._dtype = value_type
        self._key_type = 'S{}'.format(max(len(k.encode()) for k in keys))

        self._results = np.zeros(chunksize, dtype=value_type)
        self._strresult = [None] * chunksize

    def chunk_factory(self, length):
        return [None] * length

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            keys = self._keys
            results = self._results
            strresults = self._strresult
            for i in range(len(values)):
                value = keys.get(values[i], -1)
                if value == -1:
                    strresults[i] = values[i]
                else:
                    strresults[i] = ''
                results[i] = value
                # results = keys[values[i]]
            if len(values) != len(results):
                self._field.data.write_part(results[:len(values)])
                self._str_field.data.write_part(strresults[:len(values)])
            else:
                self._field.data.write_part(results)
                self._str_field.data.write_part(strresults)

    def complete(self):
        self._field.data.complete()
        self._str_field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class DateTimeImporter:
    def __init__(self, session, group, name,
                 optional=False, write_days=False, timestamp=None, chunksize=None):
        chunksize = session.chunksize if chunksize is None else chunksize
        timestamp_field_constructor(session, group, name, timestamp, chunksize)
        self._field = TimestampField(session, group, name, timestamp, write_enabled=True)
        self._results = np.zeros(chunksize , dtype='float64')
        self._optional = optional

        if optional is True:
            filter_name = '{}_set'.format(name)
            numeric_field_constructor(session, group, filter_name, 'bool',
                                      timestamp, chunksize)
            self._filter_field = NumericField(session, group, filter_name, write_enabled=True)

    def chunk_factory(self, length):
        return np.zeros(length, dtype='U32')

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            results = self._results

            for i, v in enumerate(values):
                if len(v) == 32:
                    ts = datetime.strptime(v, '%Y-%m-%d %H:%M:%S.%f%z')
                    results[i] = ts.timestamp()
                elif len(v) == 25:
                    ts = datetime.strptime(v, '%Y-%m-%d %H:%M:%S%z')
                    results[i] = ts.timestamp()
                else:
                    if self._optional is True and len(v) == 0:
                        results[i] = np.nan
                    else:
                        msg = "Date field '{}' has unexpected format '{}'"
                        raise ValueError(msg.format(self._field, v))

            self._field.data.write_part(results)

    def complete(self):
        self._field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()


class DateImporter:
    def __init__(self, session, group, name,
                 optional=False, timestamp=None, chunksize=None):
        timestamp_field_constructor(session, group, name, timestamp, chunksize)
        self._field = TimestampField(session, group, name, timestamp, write_enabled=True)
        self._results = np.zeros(chunksize, dtype='float64')

        if optional is True:
            filter_name = '{}_set'.format(name)
            numeric_field_constructor(session, group, filter_name, 'bool',
                                      timestamp, chunksize)
            self._filter_field = NumericField(session, group, filter_name, write_enabled=True)

    def chunk_factory(self, length):
        return np.zeros(length, dtype='U10')

    def write_part(self, values):
        with utils.Timer("writing {}".format(self._field.name)):
            timestamps = np.zeros(len(values), dtype=np.float64)
            for i in range(len(values)):
                value = values[i]
                if value == '':
                    timestamps[i] = np.nan
                else:
                    ts = datetime.strptime(value, '%Y-%m-%d')
                    timestamps[i] = ts.timestamp()
            self._field.data.write_part(timestamps)

    def complete(self):
        self._field.data.complete()

    def write(self, values):
        self.write_part(values)
        self.complete()



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import numpy as np

from hystore.core import dataset, utils

from hystore.processing.analytics import group_new_test_indices_by_patient, get_patients_with_old_format_tests, \
    filter_duplicate_new_tests
from hystore.processing.convert_old_assessments import ConvertOldAssessmentsV1

strformat = '%Y-%m-%d'
t_file_name = '/home/ben/covid/covid_test_export_20200601030001.csv'
a_file_name = '/home/ben/covid/assessments_export_20200601030001.csv'
cur_day = datetime.datetime.strptime("2020-05-18", strformat)


class LastActivity:
    def __init__(self):
        self.active = None

    def add(self, timestamp):
        if self.active is None:
            self.active = timestamp
        else:
            self.active = max(self.active, timestamp)

    def __repr__(self):
        return f"{self.active}"


def sort_by_test_index_count(test_indices_by_patient):
    sorted_patient_text_index_pairs = sorted([t for t in test_indices_by_patient.items()],
                                             key=lambda t: len(t[1].indices), reverse=True)
    return sorted_patient_text_index_pairs


# start
with open(t_file_name) as f:
    t_ds = dataset.Dataset(f)
t_dtss = t_ds.field_by_name('date_taken_specific')
t_patients = group_new_test_indices_by_patient(t_ds)


# get stats and print delta for old tests
# ---------------------------------------

# a_keys = ('id', 'patient_id', 'country_code', 'created_at', 'updated_at', 'version', 'had_covid_test', 'tested_covid_positive')
a_keys = ('patient_id', 'updated_at', 'had_covid_test', 'tested_covid_positive')
with open(a_file_name) as f:
    a_ds = dataset.Dataset(f, keys=a_keys,
                           show_progress_every=5000000)
                           # show_progress_every=5000000, stop_after=1000000)
print('sorting')
a_ds.sort(keys='updated_at')

# a_ids = a_ds.field_by_name('id')
a_pids = a_ds.field_by_name('patient_id')
# a_cats = a_ds.field_by_name('created_at')
a_uats = a_ds.field_by_name('updated_at')
# a_vsns = a_ds.field_by_name('version')
# a_ccs = a_ds.field_by_name('country_code')
a_hcts = a_ds.field_by_name('had_covid_test')
a_tcps = a_ds.field_by_name('tested_covid_positive')
a_patients = get_patients_with_old_format_tests(a_ds)


print('patients with old tests:', len(a_patients))
print('patients with new tests:', len(t_patients))

# build a dictionary of test counts by patient under the new system
# -----------------------------------------------------------------


s_new_patients = set(t_patients.keys())
s_old_patients = set(a_patients.keys())
s_only_in_old = s_old_patients.difference(s_new_patients)
s_only_in_new = s_new_patients.difference(s_old_patients)
s_in_both = s_old_patients.intersection(s_new_patients)
print('only in old:', len(s_only_in_old))
print('only in new:', len(s_only_in_new))
print('in_both:', len(s_in_both))

t_cleaned_patients = filter_duplicate_new_tests(t_ds, t_patients, threshold_for_diagnostic_print=10)

t_cleaned_patient_entries = sort_by_test_index_count(t_cleaned_patients)

# p_0 = t_cleaned_patient_entries[0]

print(utils.build_histogram([len(x[1].indices) for x in t_cleaned_patient_entries]))


# a_new_rows = dict({'id': list(), 'patient_id': list(), 'created_at': list(), 'updated_at': list(),
#                    'version': list(), 'country_code': list(), 'result': list(), 'mechanism': list(),
#                    'date_taken_specific': list()})
#
# # create new test rows for patients that have had only old tests
# for pk, pv in a_patients.items():
#     if pk in s_only_in_old:
#         if pv.seen_negative + pv.seen_positive == 1:
#             if pv.seen_negative:
#                 result = 'negative'
#             else:
#                 result = 'positive'
#             # take the first entry
#             i_r = pv.indices[0]
#             a_new_rows['id'].append(a_ids[i_r])
#             a_new_rows['patient_id'].append(a_pids[i_r])
#             a_new_rows['created_at'].append(a_cats[i_r])
#             a_new_rows['updated_at'].append(a_uats[i_r])
#             a_new_rows['version'].append(a_vsns[i_r])
#             a_new_rows['country_code'].append(a_vsns[i_r])
#             a_new_rows['result'].append(result)
#             a_new_rows['date_taken_specific'].append(a_cats[i_r])

# print('adapted test count:', len(a_new_rows['id']))

value_map = {
    '': 0,
    'waiting': 1,
    'no': 2,
    'yes': 3
}
fn = ConvertOldAssessmentsV1(a_ds, t_ds, value_map)
results = fn(a_patients, s_new_patients, np.zeros(a_ds.row_count(), dtype=np.uint32))

# for i in range(100):
#     print(results['id'][i], results['patient_id'][i], results['created_at'][i],
#           results['updated_at'][i],
#           results['result'][i], results['date_taken_specific'][i])
#     for i_n in a_patients[results['patient_id'][i]].indices:
#         utils.print_diagnostic_row(f"{i_n}", a_ds, i_n, keys=a_keys)
#

dest_row_count = t_ds.row_count() + len(results['id'])
destination_tests = dict()
for n, f in zip(t_ds.names_, t_ds.fields_):
    if isinstance(f, list):
        field = [None] * dest_row_count
    else:
        field = np.zeros(dest_row_count, dtype=f.dtype)
    field[:t_ds.row_count()] = f
    field[t_ds.row_count():] = results[n]



print(len(results['id']))

exit()

# visually compare tests for patients that have both old and new tests
for p in t_cleaned_patient_entries:
    if len(p[1].indices) == 4:
        print(p[0])
        #get indices for the given patient sorted by test date
        t_sorted_indices = sorted(p[1].indices, key=lambda t: t_dtss[t])
        for s in t_sorted_indices:
            utils.print_diagnostic_row(f"{p[0]}-{s}", t_ds, s, t_ds.names_)

        # sort assessments belonging to that patient by date
        a_indices = list()
        for i_r in range(a_ds.row_count()):
            if a_pids[i_r] == p[0]:
                if a_hcts[i_r] == 'True' or a_tcps[i_r] in ('no', 'yes'):
                    a_indices.append(i_r)

        a_sorted_indices = sorted(a_indices, key=lambda t: a_uats[t])
        for s in a_sorted_indices:
            utils.print_diagnostic_row(f"{p[0]}-{s}", a_ds, s, a_keys)



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv

from hystore.core import dataset, utils


# read patients in batches of n
# read assessments for those pages and output them to n


def patient_splitter(input_filename, output_filenames, sorted_indices, bucket_size):
    ch_del = ','
    ch_quote = '"'
    rows_parsed = 0
    with open(input_filename) as f_i:
        csvr = csv.reader(f_i, delimiter=ch_del, quotechar=ch_quote)

        keys = next(csvr)

        input_rows = list()
        for r in csvr:
            input_rows.append(r)

    remaining_rows = len(input_rows)

    accumulated = 0
    for ofn in output_filenames:
        with open(ofn, 'w') as f_o:
            print("writing", ofn)
            csvw = csv.writer(f_o, delimiter=ch_del, quotechar=ch_quote)

            csvw.writerow(keys)

            for i_r in range(min(bucket_size, remaining_rows)):
                ind = sorted_indices[accumulated + i_r]
                csvw.writerow(input_rows[ind])
            accumulated += bucket_size
            remaining_rows -= bucket_size


def assessment_splitter(input_filename, output_filename, assessment_buckets, bucket):
    ch_del = ','
    ch_quote = '"'
    rows_parsed = 0
    rows_written = 0
    with open(input_filename) as f_i:
        with open(output_filename, 'w') as f_o:
            csvdr = csv.DictReader(f_i, delimiter=ch_del, quotechar=ch_quote)
            keys = csvdr.fieldnames
            csvr = csv.reader(f_i, delimiter=ch_del, quotechar=ch_quote)

            csvw = csv.writer(f_o, delimiter=ch_del, quotechar=ch_quote)
            csvw.writerow(keys)

            for r in csvr:
                if rows_parsed > 0 and rows_parsed % 100000 == 0:
                    print(f"{rows_parsed} ({rows_written})")
                if assessment_buckets[rows_parsed] == bucket:
                    csvw.writerow(r)
                    rows_written += 1
                rows_parsed += 1

    print(f"complete: {rows_parsed} ({rows_written})")


def split_data(patient_data, assessment_data, bucket_size=500000, territories=None):

    with open(patient_data) as f:
        p_ds = dataset.Dataset(f, keys=('id', 'created_at'),
                               show_progress_every=500000)
                               # show_progress_every=500000, stop_after=500000)
        p_ds.sort(('created_at', 'id'))
        p_ids = p_ds.field_by_name('id')
        p_dts = p_ds.field_by_name('created_at')

    # put assessment ids into buckets
    buckets = dict()
    bucket_index = 0
    bucket_count = 0
    for i_r in range(p_ds.row_count()):
        if bucket_index == bucket_size:
            bucket_index = 0
            bucket_count += 1
        buckets[p_ids[i_r]] = bucket_count
        bucket_index += 1

    filenames = list()
    for b in range(bucket_count+1):
        destination_filename = patient_data[:-4] + f"_{b:04d}" + ".csv"
        filenames.append(destination_filename)
    print(filenames)
    sorted_indices = p_ds.index_
    del p_ds

    patient_splitter(patient_data, filenames, sorted_indices, bucket_size)

    print('buckets:', bucket_index)
    with open(assessment_data) as f:
        a_ds = dataset.Dataset(f, keys=('patient_id', 'other_symptoms'), show_progress_every=500000)

    print(utils.build_histogram(buckets.values()))

    print('associating assessments with patients')
    orphaned_assessments = 0
    a_buckets = list()
    a_pids = a_ds.field_by_name('patient_id')
    a_os = a_ds.field_by_name('other_symptoms')
    for i_r in range(a_ds.row_count()):
        if a_pids[i_r] in buckets:
            a_buckets.append(buckets[a_pids[i_r]])
        else:
            orphaned_assessments += 1
            a_buckets.append(-1)

    del a_ds
    print('orphaned_assessments:', orphaned_assessments)

    print(f'{bucket_count + 1} buckets')
    for i in range(bucket_count + 1):
        print('bucket', i)
        destination_filename = assessment_data[:-4] + f"_{i:04d}" + ".csv"
        print(destination_filename)
        # with open(assessment_data) as f:
        #     a_ds = dataset.Dataset(f, filter_fn=lambda j: a_buckets[j] == i, show_progress_every=500000)
        #
        # del a_ds
        assessment_splitter(assessment_data, destination_filename, a_buckets, i)

    print('done!')

# if __name__ == '__main__':
#     import argparse
#     parser = argparse.ArgumentParser()
#     parser.add_argument('--version', action='version', version='v0.1.9')
#     parser.add_argument('-te', '--territories', default=None,
#                         help='the territories to filter the dataset on (runs on all territories if not set)')
#     parser.add_argument('-p', '--patient_data',
#                         help='the location and name of the patient data csv file')
#     parser.add_argument('-a', '--assessment_data',
#                         help='the location and name of the assessment data csv file')
#     parser.add_argument('-b', '--bucket_size', type=int, default=500000,
#                         help='the number of patients to include in a bucket')
#
#     args = parser.parse_args()
#     if args.bucket_size < 10000:
#         print('--bucket_size cannot be less than 10000')
#         exit(-1)
#
#     utils.validate_file_exists(args.patient_data)
#     utils.validate_file_exists(args.assessment_data)
#
#     try:
#         split_data(args.patient_data, args.assessment_data, args.bucket_size, args.territories)
#     except Exception as e:
#         print(e)
#         exit(-1)



from datetime import datetime
from collections import defaultdict
import numpy as np
import h5py
import pandas as pd

from hystore.core import exporter, persistence, utils
from hystore.core.persistence import DataStore

def method_paper_summary_pipeline(ds, src_data, dest_data, first_timestamp, last_timestamp):
    s_ptnts = src_data['patients']
    s_asmts = src_data['assessments']
    filters = ds.get_or_create_group(dest_data, 'filters')
    print(s_ptnts.keys())
    print(src_data['tests'].keys())

    conditions = ('has_kidney_disease', 'has_lung_disease', 'has_heart_disease', 'has_diabetes',
                  'has_hayfever', 'has_cancer')

    symptoms = ('persistent_cough', 'fatigue', 'delirium', 'shortness_of_breath', 'fever',
                'diarrhoea', 'abdominal_pain', 'chest_pain', 'hoarse_voice', 'skipped_meals',
                'loss_of_smell')
    symptom_thresholds = {s: 2 for s in symptoms}
    symptom_thresholds['fatigue'] = 3
    symptom_thresholds['shortness_of_breath'] = 3

    intercept = -1.19015973
    weights = {'persistent_cough': 0.23186655,
               'fatigue': 0.56532346,
               'delirium': -0.12935112,
               'shortness_of_breath': 0.58273967,
               'fever': 0.16580974,
               'diarrhoea': 0.10236126,
               'abdominal_pain': -0.11204163,
               'chest_pain': -0.12318634,
               'hoarse_voice': -0.17818597,
               'skipped_meals': 0.25902482,
               'loss_of_smell': 1.82895239}

    # generating patient filter
    # -------------------------
    if 'patient_filter' not in filters.keys():
        with utils.Timer("generating patient filter", new_line=True):
            p_filter = ds.get_reader(s_ptnts['year_of_birth_valid'])[:]

            # valid age ranges
            r_ = ds.get_reader(s_ptnts['age'])[:]
            f_ = (r_ >= 18) & (r_ <= 100)
            p_filter = p_filter & f_

            # gender filter
            r_ = ds.get_reader(s_ptnts['gender'])[:]
            f_ = (r_ == 1) | (r_ == 2)
            p_filter = p_filter & f_

            # country code
            r_ = ds.get_reader(s_ptnts['country_code'])[:]
            f_ = r_ == b'GB'
            p_filter = p_filter & f_

            # no assessments
            r_ = ds.get_reader(s_ptnts['assessment_count'])[:]
            f_ = r_ > 0
            p_filter = p_filter & f_

            print("  {}, {}".format(np.count_nonzero(p_filter),
                                    np.count_nonzero(p_filter == False)))
            ds.get_numeric_writer(filters, 'patient_filter', 'bool').write(p_filter)


    # generating assessment filter
    # ----------------------------
    if 'assessment_filter' not in filters.keys():
        with utils.Timer("generating assessment filter", new_line=True):
            a_filter = np.ones(len(ds.get_reader(s_asmts['id'])), dtype=np.bool)

            # created_at in range
            r_ = ds.get_reader(s_asmts['created_at'])[:]
            f_ = (r_ >= first_timestamp) & (r_ < last_timestamp)
            a_filter = a_filter & f_

            # country code
            r_ = ds.get_reader(s_asmts['country_code'])[:]
            f_ = r_ == b'GB'
            a_filter = a_filter & f_

            with utils.Timer(f"filtering out orphaned assessments"):
                p_ids_ = ds.get_reader(s_ptnts['id'])[:]
                p_ids_ = ds.apply_filter(ds.get_reader(filters['patient_filter'])[:], p_ids_)
                a_pids_ = ds.get_reader(s_asmts['patient_id'])[:]
                f_ = persistence.foreign_key_is_in_primary_key(p_ids_, a_pids_)
            a_filter = a_filter & f_

            print("  {}, {}".format(np.count_nonzero(a_filter),
                                    np.count_nonzero(a_filter == False)))
            ds.get_numeric_writer(filters, 'assessment_filter', 'bool').write(a_filter)

    # filtering patients
    # ------------------
    if 'filtered_patients' not in dest_data.keys():
        flt_ptnts = dest_data.create_group('filtered_patients')
        with utils.Timer("filtering/flattening patient fields", new_line=True):
            p_filter = ds.get_reader(filters['patient_filter'])[:]

            r = ds.get_reader(s_ptnts['age'])
            r.get_writer(flt_ptnts, 'age').write(ds.apply_filter(p_filter, r[:]))

            for k in conditions:
                r = ds.get_reader(s_ptnts[k])
                ds.get_numeric_writer(flt_ptnts, k, 'bool').write(
                    ds.apply_filter(p_filter, r[:]) == 2)

            smoker1 = ds.get_reader(s_ptnts['is_smoker'])
            smoker2 = ds.get_reader(s_ptnts['smoker_status'])
            smoker = (smoker1[:] == 2) | (smoker2[:] == 3)
            ds.get_numeric_writer(flt_ptnts, 'smoker', 'bool').write(smoker)

            gender_ = ds.get_reader(s_ptnts['gender'])
            ds.get_numeric_writer(flt_ptnts, 'gender', 'uint8').write(
                ds.apply_filter(p_filter, gender_) - 1)
    else:
        flt_ptnts = dest_data['filtered_patients']

    # filtering assessments
    # ---------------------
    if 'filtered_assessments' not in dest_data.keys():
        flt_asmts = dest_data.create_group('filtered_assessments')
        with utils.Timer("filtering/flattening symptoms",
                         new_line=True):
            a_filter = ds.get_reader(filters['assessment_filter'])[:]
            for s in symptoms:
                r_ = ds.get_reader(s_asmts[s])[:]
                ds.get_numeric_writer(flt_asmts, s, 'bool').write(
                    ds.apply_filter(a_filter, r_) >= symptom_thresholds[s])
            a_pids = ds.get_reader(s_asmts['patient_id'])
            a_pids.get_writer(flt_asmts, 'patient_id').write(ds.apply_filter(a_filter, a_pids[:]))
    else:
        flt_asmts = dest_data['filtered_assessments']

    # predicting covid
    # ----------------
    if 'prediction' not in dest_data['filtered_assessments']:
        with utils.Timer("generating covid prediction", new_line=True):
            cumulative = np.zeros(len(ds.get_reader(flt_asmts['persistent_cough'])), dtype='float64')
            for s in symptoms:
                reader = ds.get_reader(flt_asmts[s])
                cumulative += reader[:] * weights[s]
            cumulative += intercept
            print("positive predictions", np.count_nonzero(cumulative > 0.0), len(cumulative))

            a_pids_ = ds.get_reader(flt_asmts['patient_id'])[:]
            spans = ds.get_spans(a_pids_)
            max_prediction_inds = ds.apply_spans_index_of_max(spans, cumulative)
            max_predictions = cumulative[max_prediction_inds]

            ds.get_numeric_writer(flt_asmts, 'prediction', 'float32').write(max_predictions)
            pos_filter = max_predictions > 0.0
            print("pos_filter: ", np.count_nonzero(pos_filter), len(pos_filter))

    # generating table results
    print('total_assessments:', np.count_nonzero(ds.get_reader(filters['assessment_filter'])[:]))
    subjects = np.count_nonzero(ds.get_reader(filters['patient_filter'])[:])
    genders = ds.get_reader(flt_ptnts['gender'])[:]
    predicted_c19 = np.count_nonzero(ds.get_reader(flt_asmts['prediction'])[:] > 0.0)
    age_mean = np.mean(ds.get_reader(flt_ptnts['age'])[:])
    age_std = np.std(ds.get_reader(flt_ptnts['age'])[:])
    print('subjects:', subjects)
    male = np.count_nonzero(genders)
    female = np.count_nonzero(genders == False)
    print('gender: {}:{}, {:.2%}:{:.2%}'.format(male, female,
                                              male / len(genders), female / len(genders)))
    # print('predicted covid-19:', predicted_c19)
    print('{}:'.format('predicted covid-19'),
          '{} {:.2%}'.format(predicted_c19,
                             predicted_c19 / len(ds.get_reader(flt_asmts['prediction']))))
    print('age {:.2f} ({:.2f})'.format(age_mean, age_std))
    for k in conditions + ('smoker',):
        kr_ = ds.get_reader(flt_ptnts[k])[:]
        pos = np.count_nonzero(kr_)
        print('{}:'.format(k), '{} {:.2%}'.format(pos, pos / len(kr_)))
        # print(np.count_nonzero(kr_), len(kr_))


if __name__ == '__main__':
    datastore = DataStore()
    src_file = '/home/ben/covid/ds_20200731_full.hdf5'
    dest_file = '/home/ben/covid/ds_20200731_supplement_table.hdf5'
    with h5py.File(src_file, 'r+') as src_data:
        with h5py.File(dest_file, 'w') as dest_data:
            method_paper_summary_pipeline(datastore, src_data, dest_data,
                                         datetime.strptime("2020-04-01", '%Y-%m-%d').timestamp(),
                                         datetime.strptime("2020-08-01", '%Y-%m-%d').timestamp())



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hystore.processing.covid_test import ValidateCovidTestResultsFacVersion1, ValidateCovidTestResultsFacVersion2
from hystore.processing.temperature import ValidateTemperature1
from hystore.processing.weight_height_bmi import ValidateHeight1, ValidateHeight2


class ParsingSchemaVersionError(Exception):
    pass


class ClassEntry:
    def __init__(self, key, class_definition, version_from, version_to=None):
        self.key = key
        self.class_definition = class_definition
        self.version_from = version_from
        self.version_to = version_to

    def __str__(self):
        output = 'ClassEntry(field={}, class_definition={}, version_from={}, version_to={})'
        return output.format(self.key, self.class_definition,
                             self.version_from, self.version_to)

    def __repr__(self):
        return self.__str__()


parsing_schemas = [1, 2]

class ParsingSchema:
    def __init__(self, schema_number):

        #self.parsing_schemas = [1, 2]
        self.functors = {
            'validate_weight_height_bmi': [
                ClassEntry('validate_weight_height_bmi', ValidateHeight1, 1, 2),
                ClassEntry('validate_weight_height_bmi', ValidateHeight2, 2, None)],
            'validate_temperature': [
                ClassEntry('validate_temperature', ValidateTemperature1, 1, None)],
            'clean_covid_progression': [
                ClassEntry('validate_covid_fields', ValidateCovidTestResultsFacVersion1, 1, 2),
                ClassEntry('validate_covid_fields', ValidateCovidTestResultsFacVersion2, 2, None)]
        }

        self._validate_schema_number(schema_number)

        self.class_entries = dict()
        for f in self.functors.items():
            for e in f[1]:
                if schema_number >= e.version_from and\
                  (e.version_to is None or schema_number < e.version_to):
                    self.class_entries[f[0]] = e.class_definition
                    break

    def _validate_schema_number(self, schema_number):
        if schema_number not in parsing_schemas:
            raise ParsingSchemaVersionError(
                f'{schema_number} is not a valid cleaning schema value')



import time
import numpy as np
import pandas as pd

import h5py

from hystore.core.session import Session
from hystore.core import operations as ops
from hystore.core.operations import ordered_map_to_right_left_unique
from hystore.core import fields as flds
from hystore.core import utils as utils


def generate_a_ids(length, id_base):
    a_ids = np.arange(length, dtype=np.int64) + id_base
    return a_ids


def generate_b_ids(length, id_base, mapping):
    b_ids = np.zeros(length * sum(mapping) // len(mapping), dtype=np.int64)
    cur = 0
    for i in range(length):
        count = mapping[i % 4]
        for c in range(count):
            b_ids[cur] = i + id_base
            cur += 1
    return b_ids


def generate_a_vals(length, min_val, max_val, rng):
    a_vals = rng.randint(min_val, max_val, size=length)
    return a_vals


def generate_dataset_1(length, rng):
    mapping = [0, 1, 2, 1]
    a_id_base = 100000000
    a_ids = np.arange(length, dtype=np.int64) + a_id_base
    a_vals = rng.randint(0, 100, size=length)
    b_ids = np.zeros(length * sum(mapping) // len(mapping), dtype=np.int64)
    cur = 0
    for i in range(length):
        count = mapping[i % 4]
        for c in range(count):
            b_ids[cur] = a_ids[i]
            cur += 1
    return a_ids, a_vals, b_ids


def pd_test_1(length, val_column_count):
    # a_ids, a_vals, b_ids = generate_dataset_1(length, rng)
    rng = np.random.RandomState(12345678)
    id_base = 1000000000
    mapping = [0, 1, 2, 1]
    a_ids = generate_a_ids(length, id_base)
    b_ids = generate_b_ids(length, id_base, mapping)
    a_vals = dict({'id': a_ids})
    for v in range(val_column_count):
        a_vals["vals_{}".format(v)] = generate_a_vals(length, 0, 100, rng)
    a_df = pd.DataFrame(a_vals)
    a_df.set_index('id')
    b_df = pd.DataFrame({'a_id': b_ids})
    b_df.set_index('a_id')
    t0 = time.time()
    r_df = pd.merge(left=b_df, right=a_df, left_on='a_id', right_on='id', how='left')
    # r_df = pd.merge(left=a_df, right=b_df, left_on='id', right_on='a_id', how='right')
    # r_df = b_df.join(a_df, on='a_id')
    msg = "{} elements merged with {} elements in {}s"
    print(msg.format(a_ids.size, b_ids.size, time.time() - t0))
    # for k in a_vals:
    #     if k != 'id':
    #         with utils.Timer("get data and sum"):
    #             vals = r_df[k][:]
    #             print(sum(r_df[k]))
    #         with utils.Timer("repeat get data and sum"):
    #             print(sum(vals))

def hs_test_1(length, val_column_count):
    # a_ids, a_vals, b_ids = generate_dataset_1(length)
    rng = np.random.RandomState(12345678)
    id_base = 1000000000
    mapping = [0, 1, 2, 1]
    s = Session()
    with h5py.File('/home/ben/covid/benchmarking.hdf5', 'w') as hf:
        print('creating a_ids')
        a_ids = generate_a_ids(length, id_base)
        a_ids_f = s.create_numeric(hf, 'a_ids', 'int64')
        a_ids_f.data.write(a_ids)
        del a_ids

        print('creating a_vals')
        # all_a_val_fields = list()
        for v in range(val_column_count):
            a_vals = generate_a_vals(length, 0, 100, rng)
            a_vals_f = s.create_numeric(hf, 'a_vals_{}'.format(v), 'int64')
            a_vals_f.data.write(a_vals)
            # all_a_val_fields.append(a_vals_f)
            del a_vals

        print('creating b_ids')
        b_ids = generate_b_ids(length, id_base, mapping)
        b_ids_f = s.create_numeric(hf, 'b_ids', 'int64')
        b_ids_f.data.write(b_ids)
        del b_ids

        all_b_val_fields = list()
        for v in range(val_column_count):
            b_vals_f = s.create_numeric(hf, 'b_vals_{}'.format(v), 'int32')
            all_b_val_fields.append(b_vals_f)

        a_to_b = s.create_numeric(hf, 'a_to_b', 'int64')

        all_a_val_fields = list()
        for v in range(val_column_count):
            a_vals_f = s.get(hf['a_vals_{}'.format(v)])
            all_a_val_fields.append(a_vals_f)

        print("running test")
        t0 = time.time()
        # s.ordered_left_merge(a_ids, b_ids, a_to_b, left_unique=True,
        #                      left_field_sources=(a_vals_f,), left_field_sinks=(b_vals_f,))
        s.ordered_left_merge(a_ids_f, b_ids_f, a_to_b, left_unique=True,
                             left_field_sources=tuple(all_a_val_fields),
                             left_field_sinks=tuple(all_b_val_fields))
        elapsed = time.time() - t0
        print(elapsed)
        print(all_b_val_fields[0].data[:100])

# from numba import njit
# @njit
def fast_sum(d_it):
    dsum = np.int64(0)
    for d in d_it:
        dsum += d
    return dsum


def minimal_test_1(length, count):
    rng = np.random.RandomState(12345678)
    with h5py.File('/home/ben/covid/benchmarking.hdf5', 'w') as hf:
        for c in range(count):
            vals = generate_a_vals(length, 0, 100, rng)
            with utils.Timer("writing source vals {}".format(c)):
                hf.create_dataset("vals_{}".format(c), chunks=(1<<20,), data=vals)

    with h5py.File('/home/ben/covid/benchmarking.hdf5', 'r+') as hf:
        for c in range(count):
            vname = "vals_{}".format(c)
            with utils.Timer("reading {}".format(vname)):
                vals = hf[vname][:]
            vals *= 2
            v2name = "dest_vals_{}".format(c)
            with utils.Timer("writing {}".format(v2name)):
                hf.create_dataset(v2name, chunks=(1<<20,), data=vals)


def iterator_test_1(length):
    a_ids, a_vals, b_ids = generate_dataset_1(length)
    s = Session()
    with h5py.File('/home/ben/covid/benchmarking.hdf5', 'w') as hf:
        wa_vals = s.create_numeric(hf, 'a_vals', 'int32')
        wa_vals.data.write(a_vals)

        wa_vals2 = s.get(hf['a_vals'])
        print(fast_sum(iter(ops.data_iterator(wa_vals2))))

def raw_np_test_1(length, count):
    rng = np.random.RandomState(12345678)
    for c in range(count):
        vals = generate_a_vals(length, 0, 100, rng)
        with utils.Timer("writing source vals {}".format(c)):
            np.save('/home/ben/covid/test_save/vals_{}'.format(c), vals)

    for c in range(count):
        vname = '/home/ben/covid/test_save/vals_{}.npy'.format(c)
        with utils.Timer("reading {}".format(vname)):
            vals = np.load(vname)
        vals *= 2
        v2name = '/home/ben/covid/test_save/dest_vals_{}'.format(c)
        with utils.Timer("writing {}".format(v2name)):
            np.save(vname, vals)


def read_id_from_csv(file_name, field_count):
    import csv
    with open(file_name) as f:

        rdr = csv.reader(f)
        fields = next(iter(rdr))
        if field_count == 1:
            ids = list()
            with utils.Timer("reading id from dataset"):
                for r in rdr:
                    ids.append(r[0])
        else:
            values = list()
            for _ in range(field_count):
                values.append(list())
            with utils.Timer("reading {} fields from dataset".format(field_count)):
                for r in rdr:
                    for i in range(field_count):
                        values[i].append(r[i])
                    del r

def read_fields_from_hdf5(file_name, field_count):
    fields = ('id', 'created_at', 'updated_at', 'version', 'country_code', 'reported_by_another', 'same_household_as_reporter', 'contact_additional_studies', 'year_of_birth', 'height_cm', 'weight_kg', 'gender', 'race_other', 'ethnicity', 'profile_attributes_updated_at', 'has_diabetes')
    print(len(fields))
    s = Session()
    with h5py.File(file_name, 'r') as hf:
        with utils.Timer("reading {} fields from dataset".format(field_count)):
            for f in range(field_count):
                field = s.get(hf['patients'][fields[f]])
                if isinstance(field, flds.IndexedStringField):
                    indices = field.indices[:]
                    values = field.values[:]
                else:
                    data = field.data[:]

# pd_test_1(1 << 24, 64)
# hs_test_1(1 << 30, 4)
# raw_np_test_1(1 << 29, 4)
# minimal_test_1(1 << 29, 16)
# iterator_test_1(1 << 24)

# read_id_from_csv('/home/ben/covid/patients_export_geocodes_20200830040058.csv', 1)
read_fields_from_hdf5('/home/ben/covid/ds_20200830_full.hdf5', 16)



import math
from datetime import datetime
from collections import defaultdict
import numpy as np
import h5py
import pandas as pd

from hystore.core import exporter, persistence, utils
from hystore.core.persistence import DataStore



def compare_datasets(ds, src1, src2):

    spaces1 = set(src1.keys())
    spaces2 = set(src2.keys())
    print(spaces1.difference(spaces2))
    print(spaces2.difference(spaces1))
    for sk in spaces1.intersection(spaces2):
        print(sk)
        sp1 = set(src1[sk].keys())
        sp2 = set(src2[sk].keys())
        print(sp1.difference(sp2))
        print(sp2.difference(sp1))


if __name__ == '__main__':
    datastore = DataStore()
    src_file_1 = '/home/ben/covid/ds_20200731_full.hdf5'
    src_file_2 = '/home/ben/covid/ds_20200731_sessiontest_full.hdf5'
    with h5py.File(src_file_1, 'r') as src_1:
        with h5py.File(src_file_2, 'r') as src_2:
            compare_datasets(datastore, src_1, src_2)
